'''
BSD 3-Clause License

Copyright (c) 2021, Qing Yu
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:

1. Redistributions of source code must retain the above copyright notice, this
   list of conditions and the following disclaimer.

2. Redistributions in binary form must reproduce the above copyright notice,
   this list of conditions and the following disclaimer in the documentation
   and/or other materials provided with the distribution.

3. Neither the name of the copyright holder nor the names of its
   contributors may be used to endorse or promote products derived from
   this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
'''

import geopandas as gpd
import pandas as pd
from shapely.geometry import Polygon
import math
import numpy as np
from .coordinates import getdistance
from .gisprocess import merge_polygon
import warnings


def area_to_grid(location, accuracy=500, method='rect', params='auto'):
    '''
    Generate the rectangular grids in the bounds or shape

    Parameters
    -------
    location : bounds(List) or shape(GeoDataFrame)
        Where to generate grids.
        If bounds, [lon1, lat1, lon2, lat2](WGS84), where lon1 , lat1 are the
        lower-left coordinates, lon2 , lat2 are the upper-right coordinates
        If shape, it should be GeoDataFrame
    accuracy : number
        Grid size (meter)
    method : str
        rect, tri or hexa
    params : list or dict
        Gridding parameters. 
        See https://transbigdata.readthedocs.io/en/latest/grids.html 
        for detail information about gridding parameters.
        When Gridding parameters is given, accuracy will not be used.

    Returns
    -------
    grid : GeoDataFrame
        Grid GeoDataFrame,
        LONCOL and LATCOL are the index of grids,
        HBLON and HBLAT are the center of the grids
    params : list or dict
        Gridding parameters. 
        See https://transbigdata.readthedocs.io/en/latest/grids.html 
        for detail information about gridding parameters.
    '''

    if (type(location) == list) | (type(location) == tuple):
        shape = ''
        bounds = location
    elif type(location) == gpd.geodataframe.GeoDataFrame:
        shape = location
        bounds = shape.unary_union.bounds
    else:
        raise Exception(
            'Location should be either bounds(List) or shape(GeoDataFrame)')
    lon1, lat1, lon2, lat2 = bounds
    if (lon1 > lon2) | (lat1 > lat2) | (abs(lat1) > 90) | (abs(lon1) > 180) | (
            abs(lat2) > 90) | (abs(lon2) > 180):
        raise Exception(
            'Bounds error. The input bounds should be in the order of \
                [lon1,lat1,lon2,lat2]. (lon1,lat1) is the lower left \
                    corner and (lon2,lat2) is the upper right corner.'
        )
    latStart = min(lat1, lat2)
    lonStart = min(lon1, lon2)
    deltaLon = accuracy * 360 / \
        (2 * math.pi * 6371004 * math.cos((lat1 + lat2) * math.pi / 360))
    deltaLat = accuracy * 360 / (2 * math.pi * 6371004)
    if params == 'auto':
        params = {'slon': lonStart,
                  'slat': latStart,
                  'deltalon': deltaLon,
                  'deltalat': deltaLat,
                  'theta': 0,
                  'method': method,
                  'gridsize': accuracy}
    else:
        params = convertparams(params)
        method = params['method']
    if method == 'rect':
        tmppoints = pd.DataFrame(np.array(
            np.meshgrid(
                np.arange(bounds[0] - deltaLon,
                          bounds[2] + deltaLon,
                          deltaLon/3),
                np.arange(bounds[1]-deltaLat,
                          bounds[3]+deltaLat,
                          deltaLat/3))
        ).reshape(2, -1).T,
            columns=['lon', 'lat'])
        tmppoints['LONCOL'], tmppoints['LATCOL'] = GPS_to_grid(
            tmppoints['lon'], tmppoints['lat'], params)
        tmppoints = tmppoints[['LONCOL', 'LATCOL']].drop_duplicates()
        tmppoints['geometry'] = grid_to_polygon(
            [tmppoints['LONCOL'], tmppoints['LATCOL']], params)
        tmppoints = gpd.GeoDataFrame(tmppoints)
    if (method == 'tri') | (method == 'hexa'):
        tmppoints = pd.DataFrame(
            np.array(
                np.meshgrid(
                    np.arange(bounds[0]-deltaLon,
                              bounds[2] + deltaLon,
                              deltaLon/3),
                    np.arange(bounds[1]+deltaLat,
                              bounds[3]-deltaLat,
                              deltaLat/3))
            ).reshape(2, -1).T, columns=['lon', 'lat'])
        tmppoints['loncol_1'],\
            tmppoints['loncol_2'],\
            tmppoints['loncol_3'] = GPS_to_grid(
            tmppoints['lon'], tmppoints['lat'], params)
        tmppoints = tmppoints[['loncol_1',
                               'loncol_2', 'loncol_3']].drop_duplicates()
        tmppoints['geometry'] = grid_to_polygon(
            [tmppoints['loncol_1'],
             tmppoints['loncol_2'],
             tmppoints['loncol_3']], params)
        tmppoints = gpd.GeoDataFrame(tmppoints)
    data = tmppoints
    params['gridsize'] = accuracy
    if type(shape) != gpd.geodataframe.GeoDataFrame:
        grid = gpd.GeoDataFrame(data)
        return grid, params
    else:
        data.crs = shape.crs
        data = data[data.intersects(shape.unary_union)]
        grid = gpd.GeoDataFrame(data)
        return grid, params


def area_to_params(location, accuracy=500, method='rect'):
    '''
    Generate gridding params

    Parameters
    -------
    location : bounds(List) or shape(GeoDataFrame)
        Where to generate grids.
        If bounds, [lon1, lat1, lon2, lat2](WGS84), where lon1 , lat1 are the
        lower-left coordinates, lon2 , lat2 are the upper-right coordinates
        If shape, it should be GeoDataFrame
    accuracy : number
        Grid size (meter)
    method : str
        rect, tri or hexa


    Returns
    -------
    params : list or dict
        Gridding parameters. 
        See https://transbigdata.readthedocs.io/en/latest/grids.html 
        for detail information about gridding parameters.

    '''
    if (type(location) == list) | (type(location) == tuple):
        shape = ''
        bounds = location
    elif type(location) == gpd.geodataframe.GeoDataFrame:
        shape = location
        bounds = shape.unary_union.bounds
    lon1, lat1, lon2, lat2 = bounds
    if (lon1 > lon2) | (lat1 > lat2) | (abs(lat1) > 90) | (abs(lon1) > 180) | (
            abs(lat2) > 90) | (abs(lon2) > 180):
        raise Exception(
            'Bounds error. The input bounds should be in the order \
                of [lon1,lat1,lon2,lat2]. (lon1,lat1) is the lower left \
                    corner and (lon2,lat2) is the upper right corner.'
        )
    latStart = min(lat1, lat2)
    lonStart = min(lon1, lon2)
    deltaLon = accuracy * 360 / \
        (2 * math.pi * 6371004 * math.cos((lat1 + lat2) * math.pi / 360))
    deltaLat = accuracy * 360 / (2 * math.pi * 6371004)
    params = [lonStart, latStart, deltaLon, deltaLat]
    params = convertparams(params)
    params['gridsize'] = accuracy
    params['method'] = method
    return params


def GPS_to_grid(lon, lat, params):
    '''
    Match the GPS data to the grids. The input is the columns of
    longitude, latitude, and the grids parameter. The output is the grid ID.

    Parameters
    -------
    lon : Series
        The column of longitude
    lat : Series
        The column of latitude
    params : list or dict
        Gridding parameters. 
        See https://transbigdata.readthedocs.io/en/latest/grids.html 
        for detail information about gridding parameters.

    Returns
    -------

    `Rectangle grids`
    [LONCOL,LATCOL] : list
        The two columns LONCOL and LATCOL together can specify a grid.

    `Triangle and Hexagon grids`
    [loncol_1,loncol_2,loncol_3] : list
        The index of the grid latitude. The two columns LONCOL and
        LATCOL together can specify a grid.
    '''
    params = convertparams(params)
    method = params['method']
    if method == 'rect':
        loncol, latcol = GPS_to_grids_rect(lon, lat, params)
        return [loncol, latcol]
    if method == 'tri':
        loncol_1, loncol_2, loncol_3 = GPS_to_grids_tri(lon, lat, params)
        return [loncol_1, loncol_2, loncol_3]
    if method == 'hexa':
        loncol_1, loncol_2, loncol_3 = GPS_to_grids_hexa(lon, lat, params)
        return [loncol_1, loncol_2, loncol_3]


def grid_to_centre(gridid, params):
    '''
    The center location of the grid. The input is the grid ID and
    parameters, the output is the grid center location.

    Parameters
    -------
    gridid : list
        if `Rectangle grids`
        [LONCOL,LATCOL] : Series
            The two columns LONCOL and LATCOL together can specify a grid.

        if `Triangle and Hexagon grids`
        [loncol_1,loncol_2,loncol_3] : Series
            The index of the grid latitude. The two columns LONCOL and
            LATCOL together can specify a grid.
    params : list or dict
        Gridding parameters. 
        See https://transbigdata.readthedocs.io/en/latest/grids.html 
        for detail information about gridding parameters.

    Returns
    -------
    HBLON : Series
        The longitude of the grid center
    HBLAT : Series
        The latitude of the grid center
    '''
    params = convertparams(params)
    method = params['method']
    if method == 'rect':
        loncol, latcol = gridid
        loncol = pd.Series(loncol, name='loncol')
        latcol = pd.Series(latcol, name='latcol')
        return grid_to_centre_rect(loncol, latcol, params, from_origin=False)
    if method == 'tri':
        loncol_1, loncol_2, loncol_3 = gridid
        loncol_1 = pd.Series(loncol_1, name='loncol_1')
        loncol_2 = pd.Series(loncol_2, name='loncol_2')
        loncol_3 = pd.Series(loncol_3, name='loncol_3')
        testpoint = gettripoints(loncol_1, loncol_2, loncol_3, params)
        hblon = ((testpoint['p1_x']+testpoint['p2_x'] +
                 testpoint['p3_x'])/3).values
        hblat = ((testpoint['p1_y']+testpoint['p2_y'] +
                 testpoint['p3_y'])/3).values
        return hblon, hblat
    if method == 'hexa':
        loncol_1, loncol_2, loncol_3 = gridid
        loncol_1 = pd.Series(loncol_1, name='loncol_1')
        loncol_2 = pd.Series(loncol_2, name='loncol_2')
        loncol_3 = pd.Series(loncol_3, name='loncol_3')
        lonStart = params['slon']
        latStart = params['slat']
        deltaLon = params['deltalon']
        deltaLat = params['deltalat']
        theta = params['theta']
        params = [lonStart, latStart, deltaLon, deltaLat, theta]
        x1, y1 = grid_to_centre_rect(
            loncol_1, np.zeros(len(loncol_1))-5,
            params=[params[0], params[1], params[2], params[3], theta+0])
        x2, y2 = grid_to_centre_rect(
            loncol_1, np.zeros(len(loncol_1))+5,
            params=[params[0], params[1], params[2], params[3], theta+0])
        x3, y3 = grid_to_centre_rect(
            loncol_2, np.zeros(len(loncol_2))-5,
            params=[params[0], params[1], params[2], params[3], theta+60])
        x4, y4 = grid_to_centre_rect(
            loncol_2, np.zeros(len(loncol_2))+5,
            params=[params[0], params[1], params[2], params[3], theta+60])
        x1 = pd.Series(x1)
        x2 = pd.Series(x2)
        x3 = pd.Series(x3)
        x4 = pd.Series(x4)
        y1 = pd.Series(y1)
        y2 = pd.Series(y2)
        y3 = pd.Series(y3)
        y4 = pd.Series(y4)

        def intersection(x1, y1, x2, y2, x3, y3, x4, y4):
            if np.allclose((x2[0]-x1[0]), 0):
                return x1, (x1-x3)*(y3-y4)/(x3-x4)+y3
            elif np.allclose((x3[0]-x4[0]), 0):
                return x3, (x3-x1)*(y1-y2)/(x1-x2)+y1
            x = (x1*(y2-y1)/(x2-x1)-x3/(x4-x3)*(y4-y3)+y3-y1) / \
                ((y2-y1)/(x2-x1)-(y4-y3)/(x4-x3))
            y = (x-x1)*(y2-y1)/(x2-x1)+y1
            return x, y
        hblon, hblat = intersection(x1, y1, x2, y2, x3, y3, x4, y4)
        return hblon.values, hblat.values


def grid_to_polygon(gridid, params):
    '''
    Generate the geometry column based on the grid ID.
    The input is the grid ID, the output is the geometry.
    Support rectangle, triangle and hexagon grids

    Parameters
    -------
    gridid : list
        if `Rectangle grids`
        [LONCOL,LATCOL] : Series
            The two columns LONCOL and LATCOL together can specify a grid.

        if `Triangle and Hexagon grids`
        [loncol_1,loncol_2,loncol_3] : Series
            The index of the grid latitude. The two columns LONCOL and
            LATCOL together can specify a grid.

    params : list or dict
        Gridding parameters. 
        See https://transbigdata.readthedocs.io/en/latest/grids.html 
        for detail information about gridding parameters.

    Returns
    -------
    geometry : Series
        The column of grid geographic polygon

    '''
    params = convertparams(params)
    method = params['method']

    if method == 'rect':
        loncol, latcol = gridid
        return gridid_to_polygon_rect(loncol, latcol, params)
    if method == 'tri':
        loncol_1, loncol_2, loncol_3 = gridid
        return gridid_to_polygon_tri(loncol_1, loncol_2, loncol_3, params)
    if method == 'hexa':
        loncol_1, loncol_2, loncol_3 = gridid
        return gridid_to_polygon_hexa(loncol_1, loncol_2, loncol_3, params)


def grid_to_area(data, shape, params, col=['LONCOL', 'LATCOL']):
    '''
    Input the two columns of grid ID, the geographic polygon and gridding
    paramters. The output is the grid.

    Parameters
    -------
    data : DataFrame
        Data, with two columns of grid ID
    shape : GeoDataFrame
        Geographic polygon
    params : list or dict
        Gridding parameters. 
        See https://transbigdata.readthedocs.io/en/latest/grids.html 
        for detail information about gridding parameters.
    col : List
        Column names [LONCOL,LATCOL] for rect grids or
        [loncol_1,loncol_2,loncol_3] for tri and hexa grids

    Returns
    -------
    data1 : DataFrame
        Data gridding and mapping to the corresponding geographic polygon
    '''
    params = convertparams(params)
    data1 = data[col].drop_duplicates().copy()
    data1 = gpd.GeoDataFrame(data1)
    if params['method'] == 'rect':
        LONCOL, LATCOL = col
        data1['geometry'] = gpd.points_from_xy(*grid_to_centre(
            [data1[LONCOL], data1[LATCOL]], params))
    elif params['method'] in ['tri', 'hexa']:
        loncol_1, loncol_2, loncol_3 = col
        data1['geometry'] = gpd.points_from_xy(*grid_to_centre(
            [data1[loncol_1], data1[loncol_2], data1[loncol_3]], params))
    data1 = gpd.sjoin(data1, shape)
    data1 = pd.merge(data, data1, on=col)
    return data1


def grid_to_params(grid):
    '''
    Regenerate gridding params from grid. Only support rect grids now.

    Parameters
    -------
    grid : GeoDataFrame
        grids generated by transbigdata


    Returns
    -------
    params : list or dict
        Gridding parameters. 
        See https://transbigdata.readthedocs.io/en/latest/grids.html 
        for detail information about gridding parameters.

    '''
    # 该方法未来计划增加对三角形与六边形网格支持
    grid_coord = np.array(grid['geometry'].iloc[0].exterior.coords)
    loncol = grid['LONCOL'].iloc[0]
    latcol = grid['LATCOL'].iloc[0]
    hblon = grid['geometry'].iloc[0].centroid.x
    hblat = grid['geometry'].iloc[0].centroid.y
    grid_coord = grid_coord - grid_coord[0]
    x = grid_coord[1]
    y = grid_coord[3]
    R = np.array([x, y])
    lonstart, latstart = np.array([hblon, hblat
                                   ]) - R[0, :] * loncol - R[1, :] * latcol
    deltalon = (x[0]**2 + y[0]**2).sum()**0.5
    deltalat = (x[1]**2 + y[1]**2).sum()**0.5
    theta = np.arccos(x[0] / deltalon) * 180 / np.pi
    if np.allclose(theta, 0):
        params = [lonstart, latstart, deltalon, deltalat]
    else:
        params = [lonstart, latstart, deltalon, deltalat, theta]
    params = convertparams(params)
    return params


def grid_params_optimize(data,
                         initialparams,
                         col=['uid', 'lon', 'lat'],
                         optmethod='centerdist',
                         printlog=False,
                         sample=0,
                         pop=15,
                         max_iter=50,
                         w=0.1,
                        c1=0.5,
                        c2=0.5):
    '''
    Optimize the grid params

    Parameters
    -------
    data : DataFrame
        Trajectory data
    initialparams : List
        Initial griding params
    col : List
        Column names [uid,lon,lat]
    optmethod : str
        The method to optimize: centerdist, gini, gridscount
    printlog : bool
        Whether to print detail result
    sample : int
        Sample the data as input, if 0 it will not perform sampling
    pop,max_iter,w,c1,c2:
        Params in PSO from scikit-opt

    Returns
    -------
    params_optimized : List
        Optimized params
    '''
    if sample > 0:
        trajdata = data.sample(sample)
    else:
        trajdata = data.copy()
    params = initialparams
    times = 10
    theta_lambda = 1

    params = convertparams(params)
    method = params['method']

    [uid, lon, lat] = col
    try:
        from sko.PSO import PSO
    except ImportError:
        raise ImportError(
            "Please install scikit-opt, run following code "
            "in cmd: pip install scikit-opt")

    def grids_index_gini(gpsdata, params, col=['longitude', 'latitude']):
        [lon, lat] = col
        data = gpsdata.copy()
        if params['method'] == 'rect':
            data['LONCOL'], data['LATCOL'] = GPS_to_grid(data[lon],
                                                         data[lat],
                                                         params=params)
            data['count'] = 1
            data = data.groupby(['LONCOL', 'LATCOL'])[
                'count'].sum().reset_index()
        elif (params['method'] == 'tri') | (params['method'] == 'hexa'):
            data['loncol_1'],\
                data['loncol_2'],\
                data['loncol_3'] = GPS_to_grid(data[lon],
                                               data[lat],
                                               params=params)
            data['count'] = 1
            data = data.groupby(['loncol_1', 'loncol_2', 'loncol_3'])[
                'count'].sum().reset_index()

        def GiniIndex(p):
            N = len(p)
            Q = np.mean(p)
            G = 2 / (N * (N - 1)) * (
                (N + 1) * np.sum(p) - 2 * np.sum([(N - (i + 1) + 1) * p[i]
                                                  for i in range(len(p))]))
            return G / (2 * Q)

        Gini = GiniIndex(list(data['count']))
        return Gini

    def grids_index_centerdist(gpsdata, params, col=[lon, lat]):
        [lon, lat] = col
        data = gpsdata.copy()
        data['HBLON'], data['HBLAT'] = grid_to_centre(
            GPS_to_grid(data[lon],
                        data[lat],
                        params=params),
            params=params)
        data['dist'] = getdistance(data['HBLON'], data['HBLAT'], data[lon],
                                   data[lat])
        return data['dist'].quantile(0.5)

    def grids_index_gridscount(gpsdata, params, col=[uid, lon, lat]):
        [uid, lon, lat] = col
        data = gpsdata.copy()
        if params['method'] == 'rect':
            data['LONCOL'], data['LATCOL'] = GPS_to_grid(data[lon],
                                                         data[lat],
                                                         params=params)
            return data[[
                uid, 'LONCOL', 'LATCOL'
            ]].drop_duplicates().groupby(uid)['LONCOL'].count().quantile(0.5)
        elif (params['method'] == 'tri') | (params['method'] == 'hexa'):
            data['loncol_1'],\
                data['loncol_2'],\
                data['loncol_3'] = GPS_to_grid(data[lon],
                                               data[lat],
                                               params=params)
            return data[[
                uid, 'loncol_1', 'loncol_2', 'loncol_3'
            ]].drop_duplicates().groupby(uid)['loncol_1'].count().quantile(0.5)

    if optmethod == "centerdist":

        def f_centerdist(x):
            return grids_index_centerdist(trajdata, {
                'slon': params['slon'] + x[0] / times,
                'slat': params['slat'] + x[1] / times,
                'deltalon': params['deltalon'],
                'deltalat': params['deltalat'],
                'method': params['method'],
                'theta': x[2] / theta_lambda,
            },
                col=[lon, lat])

        f = f_centerdist

    elif optmethod == "gini":

        def f_gini(x):
            return -grids_index_gini(trajdata, {
                'slon': params['slon'] + x[0] / times,
                'slat': params['slat'] + x[1] / times,
                'deltalon': params['deltalon'],
                'deltalat': params['deltalat'],
                'method': params['method'],
                'theta': x[2] / theta_lambda,
            },
                col=[lon, lat])

        f = f_gini
    elif optmethod == "gridscount":

        def f_gridscount(x):
            return grids_index_gridscount(trajdata, {
                'slon': params['slon'] + x[0] / times,
                'slat': params['slat'] + x[1] / times,
                'deltalon': params['deltalon'],
                'deltalat': params['deltalat'],
                'method': params['method'],
                'theta': x[2] / theta_lambda,
            },
                col=[uid, lon, lat])

        f = f_gridscount
    else:
        raise Exception('Method should be one of: centerdist,gini,gridscount')

    pso = PSO(func=f,
            n_dim=3,
            pop=pop, 
            max_iter=max_iter,
            lb=[0, 0, 0],
            ub=[params['deltalon'] * times, params['deltalat']
                * times, 90 * theta_lambda],
            w=w, 
            c1=c1, 
            c2=c2)
    result = pso.run()

    x = result[0]
    params_optimized = {
        'slon': params['slon'] + x[0] / times,
        'slat': params['slat'] + x[1] / times,
        'deltalon': params['deltalon'],
        'deltalat': params['deltalat'],
        'theta': x[2] / theta_lambda,
        'method': method
    }

    if printlog:
        print('Optimized index ' + optmethod + ':', f(result[0]))
        print('Optimized gridding params:', params_optimized)
        import matplotlib.pyplot as plt
        plt.figure(1, (14, 5), dpi=300)
        ax = plt.subplot(121)
        plt.plot(pso.gbest_y_hist)
        plt.ylabel('Cost')
        plt.xlabel('Iters')
        plt.title('Optimize cost')
        # 生成点的geodataframe
        trajdata['geometry'] = gpd.points_from_xy(trajdata[lon], trajdata[lat])
        trajdata = gpd.GeoDataFrame(trajdata)
        # 效果
        if method == 'rect':
            trajdata['LONCOL'], trajdata['LATCOL'] = GPS_to_grid(
                trajdata[lon], trajdata[lat], params=params_optimized)
            grids = trajdata.drop_duplicates(
                subset=['LONCOL', 'LATCOL']).copy()
            grids['geometry'] = grid_to_polygon([grids['LONCOL'],
                                                grids['LATCOL']],
                                                params=params_optimized)
        elif (method == 'tri') | (method == 'hexa'):
            result = GPS_to_grid(
                trajdata[lon], trajdata[lat], params=params_optimized)
            trajdata['loncol_1'], trajdata['loncol_2'], trajdata['loncol_3'] = result
            grids = trajdata.drop_duplicates(
                subset=['loncol_1', 'loncol_2', 'loncol_3']).copy()
            grids['geometry'] = grid_to_polygon([grids['loncol_1'],
                                                grids['loncol_2'],
                                                grids['loncol_3']],
                                                params=params_optimized)
        grids = gpd.GeoDataFrame(grids)

        ax = plt.subplot(122)
        plt.sca(ax)
        plt.axis('off')
        trajdata.plot(ax=ax, markersize=0.3)
        grids.plot(lw=1,
                   edgecolor=(0.8, 0.8, 0.8, 1),
                   facecolor=(0, 0, 0, 0.05),
                   ax=ax)
        plt.title('Gridding result')
        plt.show()
    return params_optimized


'''
Utils
'''


def grid_to_centre_rect(loncol, latcol, params, from_origin=False):
    '''
    The center location of the grid for rect grids. The input is the
    grid ID and parameters, the output is the grid center location.

    Parameters
    -------
    LONCOL : Series
        The index of the grid longitude. The two columns LONCOL and
        LATCOL together can specify a grid.
    LATCOL : Series
        The index of the grid latitude. The two columns LONCOL and
        LATCOL together can specify a grid.
    params : List
        Gridding parameters (lonStart, latStart, deltaLon, deltaLat) or
        (lonStart, latStart, deltaLon, deltaLat, theta).
        lonStart and latStart are the lower-left coordinates;
        deltaLon, deltaLat are the length and width of a single grid;
        theta is the angle of the grid, it will be 0 if not given.
        When Gridding parameters is given, accuracy will not be used.
    from_origin : bool
        If True, lonStart and latStart are the lower left of the first
        grid.
        If False, lonStart and latStart are the center of the first
        grid.

    Returns
    -------
    HBLON : Series
        The longitude of the grid center
    HBLAT : Series
        The latitude of the grid center
    '''
    params = convertparams(params)
    lonStart = params['slon']
    latStart = params['slat']
    deltaLon = params['deltalon']
    deltaLat = params['deltalat']
    theta = params['theta']

    loncol = pd.Series(loncol)
    latcol = pd.Series(latcol)
    costheta = np.cos(theta * np.pi / 180)
    sintheta = np.sin(theta * np.pi / 180)
    R = np.array([[costheta * deltaLon, -sintheta * deltaLat],
                  [sintheta * deltaLon, costheta * deltaLat]])
    if from_origin:
        hblonhblat = np.dot(np.array([loncol.values, latcol.values]).T,
                            R) + np.array([lonStart, latStart]) - (
                                R[0, :] / 2 + R[1, :] / 2)
    else:
        hblonhblat = np.dot(np.array([loncol.values, latcol.values]).T,
                            R) + np.array([lonStart, latStart])
    hblon = hblonhblat[:, 0]
    hblat = hblonhblat[:, 1]
    if len(hblon) == 1:
        hblon = hblon[0]
        hblat = hblat[0]
    return hblon, hblat


def GPS_to_grids_rect(lon, lat, params, from_origin=False):
    '''
    Match the GPS data to the grids. The input is the columns of
    longitude, latitude, and the grids parameter. The output is the grid ID.

    Parameters
    -------
    lon : Series
        The column of longitude
    lat : Series
        The column of latitude
    params : List
        Gridding parameters (lonStart, latStart, deltaLon, deltaLat) or
        (lonStart, latStart, deltaLon, deltaLat, theta).
        lonStart and latStart are the lower-left coordinates;
        deltaLon, deltaLat are the length and width of a single grid;
        theta is the angle of the grid, it will be 0 if not given.
        When Gridding parameters is given, accuracy will not be used.
    from_origin : bool
        If True, lonStart and latStart are the lower left of the first
        grid.
        If False, lonStart and latStart are the center of the first
        grid.

    Returns
    -------
    LONCOL : Series
        The index of the grid longitude. The two columns LONCOL and
        LATCOL together can specify a grid.
    LATCOL : Series
        The index of the grid latitude. The two columns LONCOL and
        LATCOL together can specify a grid.
    '''
    params = convertparams(params)
    lonStart = params['slon']
    latStart = params['slat']
    deltaLon = params['deltalon']
    deltaLat = params['deltalat']
    theta = params['theta']

    lon = pd.Series(lon)
    lat = pd.Series(lat)
    costheta = np.cos(theta * np.pi / 180)
    sintheta = np.sin(theta * np.pi / 180)
    R = np.array([[costheta * deltaLon, -sintheta * deltaLat],
                  [sintheta * deltaLon, costheta * deltaLat]])
    coords = np.array([lon, lat]).T
    if from_origin:
        coords = coords - (np.array([lonStart, latStart]))
    else:
        coords = coords - (np.array([lonStart, latStart]) - R[0, :] / 2 -
                           R[1, :] / 2)
    res = np.floor(np.dot(coords, np.linalg.inv(R)))
    loncol = res[:, 0].astype(int)
    latcol = res[:, 1].astype(int)
    if len(loncol) == 1:
        loncol = loncol[0]
        latcol = latcol[0]
    return loncol, latcol


def GPS_to_grids_tri(lon, lat, params):
    '''
    Match the GPS data to the grids. The input is the columns of
    longitude, latitude, and the grids parameter. The output is the grid ID.

    Parameters
    -------
    lon : Series
        The column of longitude
    lat : Series
        The column of latitude
    params : List
        Gridding parameters (lonStart, latStart, deltaLon, deltaLat) or
        (lonStart, latStart, deltaLon, deltaLat, theta).
        lonStart and latStart are the lower-left coordinates;
        deltaLon, deltaLat are the length and width of a single grid;
        theta is the angle of the grid, it will be 0 if not given.
        When Gridding parameters is given, accuracy will not be used.

    Returns
    -------
    loncol_1,loncol_2,loncol_3 : Series
        The index of the triangle grid.
    '''
    lon = pd.Series(lon, name='lon')
    lat = pd.Series(lat, name='lat')

    params = convertparams(params)
    lonStart = params['slon']
    latStart = params['slat']
    deltaLon = params['deltalon']
    deltaLat = params['deltalat']
    theta = params['theta']
    params = [lonStart, latStart, deltaLon, deltaLat, theta]

    loncol_1, _ = GPS_to_grids_rect(lon, lat,
                                    params=[*params[:4], theta+0],
                                    from_origin=True)
    loncol_2, _ = GPS_to_grids_rect(lon, lat,
                                    params=[*params[:4], theta+60],
                                    from_origin=True)
    loncol_3, _ = GPS_to_grids_rect(lon, lat,
                                    params=[*params[:4], theta+120],
                                    from_origin=True)
    return loncol_1, loncol_2, loncol_3


def GPS_to_grids_hexa(lon, lat, params):
    '''
    Match the GPS data to the grids. The input is the columns of
    longitude, latitude, and the grids parameter. The output is the grid ID.

    Parameters
    -------
    lon : Series
        The column of longitude
    lat : Series
        The column of latitude
    params : List
        Gridding parameters (lonStart, latStart, deltaLon, deltaLat) or
        (lonStart, latStart, deltaLon, deltaLat, theta).
        lonStart and latStart are the lower-left coordinates;
        deltaLon, deltaLat are the length and width of a single grid;
        theta is the angle of the grid, it will be 0 if not given.
        When Gridding parameters is given, accuracy will not be used.

    Returns
    -------
    loncol_1,loncol_2,loncol_3 : Series
        The index of the hexagon grid.
    '''
    lon = pd.Series(lon, name='lon')
    lat = pd.Series(lat, name='lat')
    params = convertparams(params)
    lonStart = params['slon']
    latStart = params['slat']
    deltaLon = params['deltalon']
    deltaLat = params['deltalat']
    theta = params['theta']
    params = [lonStart, latStart, deltaLon, deltaLat, theta]

    loncol_1, _ = GPS_to_grids_rect(lon, lat,
                                    params=[*params[:4], theta+0],
                                    from_origin=True)
    loncol_2, _ = GPS_to_grids_rect(lon, lat,
                                    params=[*params[:4], theta+60],
                                    from_origin=True)
    loncol_3, _ = GPS_to_grids_rect(lon, lat,
                                    params=[*params[:4], theta+120],
                                    from_origin=True)
    tmp = pd.DataFrame()
    tmp['lon'] = lon
    tmp['lat'] = lat
    tmp['loncol_1'] = loncol_1
    tmp['loncol_2'] = loncol_2
    tmp['loncol_3'] = loncol_3
    tmp['id'] = range(len(tmp))
    tmp1 = tmp[['loncol_1', 'loncol_2', 'loncol_3']].drop_duplicates()
    loncol_1 = tmp1['loncol_1']
    loncol_2 = tmp1['loncol_2']
    loncol_3 = tmp1['loncol_3']
    center = pd.concat(
        [pd.DataFrame(
            np.array(
                [loncol_1+1, loncol_2+1, loncol_3+1,
                 loncol_1, loncol_2, loncol_3]).T),
         pd.DataFrame(
            np.array(
                [loncol_1+1, loncol_2+1, loncol_3,
                 loncol_1, loncol_2, loncol_3]).T),
         pd.DataFrame(
            np.array(
                [loncol_1+1, loncol_2, loncol_3,
                 loncol_1, loncol_2, loncol_3]).T),
         pd.DataFrame(
            np.array(
                [loncol_1, loncol_2+1, loncol_3+1,
                 loncol_1, loncol_2, loncol_3]).T),
         pd.DataFrame(
            np.array(
                [loncol_1, loncol_2, loncol_3+1,
                 loncol_1, loncol_2, loncol_3]).T),
         pd.DataFrame(
             np.array(
                 [loncol_1, loncol_2, loncol_3,
                  loncol_1, loncol_2, loncol_3]).T)]).drop_duplicates()
    center.columns = ['i', 'j', 'k', 'loncol_1', 'loncol_2', 'loncol_3']
    center = center[((((center['i']-1) % 3) == 0) &
                     (((center['j']-1) % 3) == 0) &
                     (((center['k']) % 3) == 0)) |
                    ((((center['i']-2) % 3) == 0) &
                     (((center['j']) % 3) == 0) &
                     (((center['k']+2) % 3) == 0)) |
                    ((((center['i']) % 3) == 0) &
                     (((center['j']+1) % 3) == 0) &
                     (((center['k']+1) % 3) == 0))
                    ]
    tmp = pd.merge(tmp, center, on=['loncol_1', 'loncol_2', 'loncol_3'])
    tmp = tmp.sort_values(by='id')
    loncol_1 = tmp['i'].values
    loncol_2 = tmp['j'].values
    loncol_3 = tmp['k'].values
    return loncol_1, loncol_2, loncol_3


def gridid_to_polygon_rect(loncol, latcol, params):
    '''
    Generate the geometry column based on the grid ID.
    The input is the grid ID, the output is the geometry.

    Parameters
    -------
    LONCOL : Series
        The index of the grid longitude.
        The two columns LONCOL and LATCOL together can specify a grid.
    LATCOL : Series
        The index of the grid latitude.
        The two columns LONCOL and LATCOL together can specify a grid.
    params : List
        Gridding parameters (lonStart, latStart, deltaLon, deltaLat) or
        (lonStart, latStart, deltaLon, deltaLat, theta).
        lonStart and latStart are the lower-left coordinates;
        deltaLon, deltaLat are the length and width of a single grid;
        theta is the angle of the grid, it will be 0 if not given.
        When Gridding parameters is given, accuracy will not be used.

    Returns
    -------
    geometry : Series
        The column of grid geographic polygon
    '''
    params = convertparams(params)
    lonStart = params['slon']
    latStart = params['slat']
    deltaLon = params['deltalon']
    deltaLat = params['deltalat']
    theta = params['theta']
    loncol = pd.Series(loncol)
    latcol = pd.Series(latcol)
    costheta = np.cos(theta * np.pi / 180)
    sintheta = np.sin(theta * np.pi / 180)
    R = np.array([[costheta * deltaLon, -sintheta * deltaLat],
                  [sintheta * deltaLon, costheta * deltaLat]])
    res_a = np.array([loncol.values - 0.5, latcol.values - 0.5]).T
    res_b = np.array([loncol.values + 0.5, latcol.values - 0.5]).T
    res_c = np.array([loncol.values + 0.5, latcol.values + 0.5]).T
    res_d = np.array([loncol.values - 0.5, latcol.values + 0.5]).T
    hblonhblat_a = np.dot(res_a, R) + np.array([lonStart, latStart])
    hblonhblat_b = np.dot(res_b, R) + np.array([lonStart, latStart])
    hblonhblat_c = np.dot(res_c, R) + np.array([lonStart, latStart])
    hblonhblat_d = np.dot(res_d, R) + np.array([lonStart, latStart])
    a = hblonhblat_a
    b = hblonhblat_b
    c = hblonhblat_c
    d = hblonhblat_d
    from shapely.geometry import Polygon
    return [Polygon([a[i], b[i], c[i], d[i], a[i]]) for i in range(len(a))]


def gettripoints(loncol_1, loncol_2, loncol_3, params):
    # Get triangle vertix coords from gridid and params
    params = convertparams(params)
    lonStart = params['slon']
    latStart = params['slat']
    deltaLon = params['deltalon']
    deltaLat = params['deltalat']
    theta = params['theta']
    params = [lonStart, latStart, deltaLon, deltaLat, theta]

    flag = (loncol_1+loncol_2+loncol_3)
    x1, y1 = grid_to_centre_rect(
        loncol_1+flag % 2, np.zeros(len(loncol_1))-5,
        params=[params[0], params[1], params[2], params[3], theta+0])
    x2, y2 = grid_to_centre_rect(
        loncol_1+flag % 2, np.zeros(len(loncol_1))+5,
        params=[params[0], params[1], params[2], params[3], theta+0])
    x3, y3 = grid_to_centre_rect(
        loncol_2+(flag+1) % 2, np.zeros(len(loncol_2))-5,
        params=[params[0], params[1], params[2], params[3], theta+60])
    x4, y4 = grid_to_centre_rect(
        loncol_2+(flag+1) % 2, np.zeros(len(loncol_2))+5,
        params=[params[0], params[1], params[2], params[3], theta+60])
    x5, y5 = grid_to_centre_rect(
        loncol_3+flag % 2, np.zeros(len(loncol_3))-5,
        params=[params[0], params[1], params[2], params[3], theta+120])
    x6, y6 = grid_to_centre_rect(
        loncol_3+flag % 2, np.zeros(len(loncol_3))+5,
        params=[params[0], params[1], params[2], params[3], theta+120])
    x1 = pd.Series(x1)
    x2 = pd.Series(x2)
    x3 = pd.Series(x3)
    x4 = pd.Series(x4)
    x5 = pd.Series(x5)
    x6 = pd.Series(x6)
    y1 = pd.Series(y1)
    y2 = pd.Series(y2)
    y3 = pd.Series(y3)
    y4 = pd.Series(y4)
    y5 = pd.Series(y5)
    y6 = pd.Series(y6)

    def intersection(x1, y1, x2, y2, x3, y3, x4, y4):
        if np.allclose((x2[0]-x1[0]), 0):
            return x1, (x1-x3)*(y3-y4)/(x3-x4)+y3
        elif np.allclose((x3[0]-x4[0]), 0):
            return x3, (x3-x1)*(y1-y2)/(x1-x2)+y1
        x = (x1*(y2-y1)/(x2-x1)-x3/(x4-x3)*(y4-y3)+y3-y1) / \
            ((y2-y1)/(x2-x1)-(y4-y3)/(x4-x3))
        y = (x-x1)*(y2-y1)/(x2-x1)+y1
        return x, y

    testpoint = pd.DataFrame()
    testpoint['p1_x'], testpoint['p1_y'] = intersection(
        x1, y1, x2, y2, x3, y3, x4, y4)
    testpoint['p2_x'], testpoint['p2_y'] = intersection(
        x5, y5, x6, y6, x3, y3, x4, y4)
    testpoint['p3_x'], testpoint['p3_y'] = intersection(
        x1, y1, x2, y2, x5, y5, x6, y6)
    return testpoint.round(6)


def gridid_to_polygon_tri(loncol_1, loncol_2, loncol_3, params):
    '''
    Generate the geometry column based on the grid ID.
    The input is the grid ID, the output is the geometry.

    Parameters
    -------
    loncol_1,loncol_2,loncol_3 : Series
        The index of the triangle grid.
    params : List
        Gridding parameters (lonStart, latStart, deltaLon, deltaLat) or
        (lonStart, latStart, deltaLon, deltaLat, theta).
        lonStart and latStart are the lower-left coordinates;
        deltaLon, deltaLat are the length and width of a single grid;
        theta is the angle of the grid, it will be 0 if not given.
        When Gridding parameters is given, accuracy will not be used.

    Returns
    -------
    geometry : Series
        The column of grid geographic polygon
    '''
    params = convertparams(params)
    lonStart = params['slon']
    latStart = params['slat']
    deltaLon = params['deltalon']
    deltaLat = params['deltalat']
    theta = params['theta']
    params = [lonStart, latStart, deltaLon, deltaLat, theta]

    loncol_1 = pd.Series(loncol_1, name='i')
    loncol_2 = pd.Series(loncol_2, name='j')
    loncol_3 = pd.Series(loncol_3, name='k')

    testpoint = gettripoints(loncol_1, loncol_2, loncol_3, params)

    geometry = testpoint.apply(
        lambda r: Polygon(np.array([[r['p1_x'], r['p1_y']],
                                    [r['p2_x'], r['p2_y']],
                                    [r['p3_x'], r['p3_y']]]).round(6)), axis=1)
    return list(geometry)


def gridid_to_polygon_hexa(loncol_1, loncol_2, loncol_3, params):
    '''
    Generate the geometry column based on the grid ID.
    The input is the grid ID, the output is the geometry.

    Parameters
    -------
    loncol_1,loncol_2,loncol_3 : Series
        The index of the hexagon grid.
    params : List
        Gridding parameters (lonStart, latStart, deltaLon, deltaLat) or
        (lonStart, latStart, deltaLon, deltaLat, theta).
        lonStart and latStart are the lower-left coordinates;
        deltaLon, deltaLat are the length and width of a single grid;
        theta is the angle of the grid, it will be 0 if not given.
        When Gridding parameters is given, accuracy will not be used.

    Returns
    -------
    geometry : Series
        The column of grid geographic polygon
    '''
    i = pd.Series(loncol_1, name='i')
    j = pd.Series(loncol_2, name='j')
    k = pd.Series(loncol_3, name='k')
    gridid = pd.DataFrame(np.array([i, j, k]).T, columns=['i', 'j', 'k'])
    tmp = pd.concat([pd.DataFrame(np.array([i-1, j-1, k-1, i, j, k]).T),
                     pd.DataFrame(np.array([i-1, j-1, k, i, j, k]).T),
                     pd.DataFrame(np.array([i-1, j, k, i, j, k]).T),
                     pd.DataFrame(np.array([i, j-1, k-1, i, j, k]).T),
                     pd.DataFrame(np.array([i, j, k-1, i, j, k]).T),
                     pd.DataFrame(np.array([i, j, k, i, j, k]).T)])
    tmp.columns = ['loncol_1', 'loncol_2', 'loncol_3', 'i', 'j', 'k']
    tmp['geometry'] = list(
        gettripoints(
            tmp['loncol_1'],
            tmp['loncol_2'],
            tmp['loncol_3'], params).apply(
            lambda r: Polygon(
                np.array([[r['p1_x'], r['p1_y']],
                          [r['p2_x'], r['p2_y']],
                          [r['p3_x'], r['p3_y']]]).round(6)), axis=1))
    tmp = gpd.GeoDataFrame(tmp)
    tmp['gridid'] = tmp['i'].astype(
        'str')+','+tmp['j'].astype('str')+','+tmp['k'].astype('str')
    tmp1 = merge_polygon(tmp, 'gridid')
    geometry = pd.merge(gridid, pd.merge(
        tmp1, tmp[['gridid', 'i', 'j', 'k']].drop_duplicates()))['geometry']
    return list(geometry)


def convertparams(params):
    # Convertparams from list to dict
    if (type(params) == list) | (type(params) == tuple):
        if len(params) == 4:
            lonStart, latStart, deltaLon, deltaLat = params
            theta = 0
            method = 'rect'
        elif len(params) == 5:
            lonStart, latStart, deltaLon, deltaLat, theta = params
            method = 'rect'
        elif len(params) == 6:
            lonStart, latStart, deltaLon, deltaLat, theta, method = params
        dicparams = dict()
        dicparams['slon'] = lonStart
        dicparams['slat'] = latStart
        dicparams['deltalon'] = deltaLon
        dicparams['deltalat'] = deltaLat
        dicparams['theta'] = theta
        dicparams['method'] = method
    else:
        dicparams = params
        if 'theta' not in dicparams:
            dicparams['theta'] = 0
        if 'method' not in dicparams:
            dicparams['method'] = 'rect'
    if dicparams['method'] not in ['rect', 'tri', 'hexa']:
        raise ValueError('Method should be `rect`,`tri` or `hexa`')
    return dicparams


'''
Old namespace
'''


def regenerate_params(*args, **kwargs):
    warnings.warn("This method is renamed as transbigdata.grid_to_params")
    return grid_to_params(*args, **kwargs)


def grid_params(*args, **kwargs):
    warnings.warn("This method is renamed as transbigdata.area_to_params")
    return area_to_params(*args, **kwargs)


def GPS_to_grids(*args, **kwargs):
    warnings.warn("This method is renamed as transbigdata.GPS_to_grid")
    return GPS_to_grid(*args, **kwargs)


def rect_grids(*args, **kwargs):
    warnings.warn("This method is renamed as transbigdata.area_to_grid")
    return area_to_grid(*args, **kwargs)


def gridid_sjoin_shape(*args, **kwargs):
    warnings.warn("This method is renamed as transbigdata.grid_to_area")
    return grid_to_area(*args, **kwargs)


def grids_centre(loncol, latcol, params):
    warnings.warn("This method is renamed as transbigdata.grid_to_centre")
    return grid_to_centre([loncol, latcol], params)


def gridid_to_polygon(loncol, latcol, params):
    warnings.warn("This method is renamed as transbigdata.grid_to_polygon")
    return grid_to_polygon([loncol, latcol], params)


'''
Geohash
'''

__base32 = '0123456789bcdefghjkmnpqrstuvwxyz'
__decodemap = {}
for i in range(len(__base32)):
    __decodemap[__base32[i]] = i
del i


def decode_exactly(geohash):
    lat_interval, lon_interval = (-90.0, 90.0), (-180.0, 180.0)
    lat_err, lon_err = 90.0, 180.0
    is_even = True
    for c in geohash:
        cd = __decodemap[c]
        for mask in [16, 8, 4, 2, 1]:
            if is_even:
                lon_err /= 2
                if cd & mask:
                    lon_interval = (
                        (lon_interval[0]+lon_interval[1])/2, lon_interval[1])
                else:
                    lon_interval = (
                        lon_interval[0], (lon_interval[0]+lon_interval[1])/2)
            else:
                lat_err /= 2
                if cd & mask:
                    lat_interval = (
                        (lat_interval[0]+lat_interval[1])/2, lat_interval[1])
                else:
                    lat_interval = (
                        lat_interval[0], (lat_interval[0]+lat_interval[1])/2)
            is_even = not is_even
    lat = (lat_interval[0] + lat_interval[1]) / 2
    lon = (lon_interval[0] + lon_interval[1]) / 2
    return lon, lat,  lon_err, lat_err


def decode(geohash):
    lon, lat,  _, _ = decode_exactly(geohash)
    return lon, lat


def encode(longitude, latitude, precision=12):
    lat_interval, lon_interval = (-90.0, 90.0), (-180.0, 180.0)
    geohash = []
    bits = [16, 8, 4, 2, 1]
    bit = 0
    ch = 0
    even = True
    i = 1
    while len(geohash) < precision:
        if even:
            mid = (lon_interval[0] + lon_interval[1]) / 2
            if longitude > mid:
                ch |= bits[bit]
                lon_interval = (mid, lon_interval[1])
            else:
                lon_interval = (lon_interval[0], mid)
        else:
            mid = (lat_interval[0] + lat_interval[1]) / 2
            if latitude > mid:
                ch |= bits[bit]
                lat_interval = (mid, lat_interval[1])
            else:
                lat_interval = (lat_interval[0], mid)
        even = not even
        if bit < 4:
            bit += 1
        else:
            geohash += __base32[ch]
            bit = 0
            ch = 0
        i += 1
    return ''.join(geohash)


def geohash_encode(lon, lat, precision=12):
    '''
    Input latitude and longitude and precision, and encode geohash code

    Parameters
    -------
    lon : Series
        longitude Series
    lat : Series
        latitude Series
    precision : number
        geohash precision

    Returns
    -------
    geohash : Series
        encoded geohash Series
    '''
    tmp = pd.DataFrame()
    tmp['lon'] = lon
    tmp['lat'] = lat
    geohash = tmp.apply(lambda r: encode(
        r['lon'], r['lat'], precision), axis=1)
    return geohash


def geohash_decode(geohash):
    '''
    Decode geohash code

    Parameters
    -------
    geohash : Series
        encoded geohash Series

    Returns
    -------
    lon : Series
        decoded longitude Series
    lat : Series
        decoded latitude Series
    '''
    lonslats = geohash.apply(decode)
    lon = lonslats.apply(lambda r: r[0])
    lat = lonslats.apply(lambda r: r[1])
    return lon, lat


def geohash_togrid(geohash):
    '''
    Input geohash code to generate geohash grid cell

    Parameters
    -------
    geohash : Series
        encoded geohash Series

    Returns
    -------
    poly : Series
        grid cell polygon for geohash
    '''
    lonslats = geohash.apply(decode_exactly)

    def topoly(r):
        (lon, lat, lon_err, lat_err) = r
        from shapely.geometry import Polygon
        return Polygon([[lon-lon_err, lat-lat_err],
                        [lon-lon_err, lat+lat_err],
                        [lon+lon_err, lat+lat_err],
                        [lon+lon_err, lat-lat_err],
                        [lon-lon_err, lat-lat_err],
                        ])
    poly = lonslats.apply(topoly)
    return poly



"""
`TransBigData`: A Python package develop for transportation spatio-temporal big
data processing, analysis and visualization.

BSD 3-Clause License

Copyright (c) 2021, Qing Yu
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:

1. Redistributions of source code must retain the above copyright notice, this
   list of conditions and the following disclaimer.

2. Redistributions in binary form must reproduce the above copyright notice,
   this list of conditions and the following disclaimer in the documentation
   and/or other materials provided with the distribution.

3. Neither the name of the copyright holder nor the names of its
   contributors may be used to endorse or promote products derived from
   this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
"""

__version__ = '0.4.10'
__author__ = 'Qing Yu <qingyu0815@foxmail.com>'

# module level doc-string
__doc__ = """
TransBigData - A Python package develop for transportation spatio-temporal big
data processing, analysis and visualization.
=====================================================================

`TransBigData` is a Python package developed for transportation spatio-temporal
big data processing, analysis and visualization. `TransBigData` provides fast
and concise methods for processing common transportation spatio-temporal big
data such as Taxi GPS data, bicycle sharing data and bus GPS data.
`TransBigData` provides a variety of processing methods for each stage of
transportation spatio-temporal big data analysis. The code with `TransBigData`
is clean, efficient, flexible, and easy to use, allowing complex data tasks to
be achieved withconcise code.

For some specific types of data, `TransBigData` also provides targeted tools
for specific needs, such as extraction of Origin and Destination(OD) of taxi
trips from taxi GPS data and identification of arrival and departure
information from bus GPS data. The latest stable release of the software can
be installed via pip and full documentation can be found at
https://transbigdata.readthedocs.io/en/latest/.

Main Features
-------------

Currently, `TransBigData` mainly provides the following methods:

  - Data Quality: Provides methods to quickly obtain the general information of
  the dataset, including the data amount the time period and the sampling
  interval.
  - Data Preprocess: Provides methods to clean multiple types of data error.
  - Data Gridding: Provides methods to generate multiple types of geographic
  grids (Rectangular grids, Hexagonal grids) in the research area. Provides
  fast algorithms to map GPS data to the generated grids.
  - Data Aggregating: Provides methods to aggregate GPS data and OD data into
  geographic polygon.
  - Data Visualization: Built-in visualization capabilities leverage the
  visualization package keplergl to interactively visualize data on Jupyter
  notebook with simple code.
  - Trajectory Processing: Provides methods to process trajectory data,
  including generating trajectory linestring from GPS points, and trajectory
  densification, etc.
  - Basemap Loading: Provides methods to display Mapbox basemap on matplotlib
  figures
"""

from .plotmap import (
    plot_map,
    plotscale,
    set_mapboxtoken,
    set_imgsavepath
)
from .coordinates import (
    gcj02tobd09,
    bd09togcj02,
    wgs84togcj02,
    gcj02towgs84,
    wgs84tobd09,
    bd09towgs84,
    bd09mctobd09,
    getdistance,
    transform_shape
)
from .grids import (
    # new
    area_to_grid,
    area_to_params,
    GPS_to_grid,
    grid_to_centre,
    grid_to_polygon,
    grid_to_area,
    grid_to_params,
    grid_params_optimize,
    # old
    rect_grids,
    grid_params,
    grids_centre,
    gridid_sjoin_shape,
    regenerate_params,
    geohash_encode,
    geohash_decode,
    geohash_togrid,
    GPS_to_grids,
    GPS_to_grids_rect,
    GPS_to_grids_tri,
    GPS_to_grids_hexa,
    gridid_to_polygon,
    gridid_to_polygon_rect,
    gridid_to_polygon_tri,
    gridid_to_polygon_hexa
)
from .gisprocess import (
    ckdnearest,
    ckdnearest_point,
    ckdnearest_line,
    splitline_with_length,
    merge_polygon,
    polyon_exterior,
    ellipse_params,
    ellipse_plot
)
from .odprocess import (
    odagg_grid,
    odagg_shape,
    tolinewitharrow
)
from .preprocess import (
    clean_same,
    clean_drift,
    clean_outofbounds,
    clean_outofshape,
    clean_traj,
    dataagg,
    id_reindex_disgap,
    id_reindex
)
from .bikedata import (
    bikedata_to_od
)
from .taxigps import (
    clean_taxi_status,
    taxigps_to_od,
    taxigps_traj_point
)
from .mobilephonedata import (
    #new
    mobile_stay_move,
    mobile_plot_activity,
    mobile_stay_dutation,
    mobile_identify_home,
    mobile_identify_work,
    #old    
    plot_activity,
    traj_stay_move,
    )
from .traj import (
    traj_densify,
    traj_sparsify,
    points_to_traj,
    dumpjson
)
from .quality import (
    sample_duration,
    data_summary
)
from .busgps import (
    busgps_arriveinfo,
    busgps_onewaytime
)
from .crawler import (
    getadmin,
    getbusdata,
    get_isochrone_amap,
    get_isochrone_mapbox,
)
from .metro import (
    split_subwayline,
    metro_network,
    get_path_traveltime,
    get_shortest_path,
    get_k_shortest_paths
)
from .visualization import (
    visualization_trip,
    visualization_od,
    visualization_data
)



'''
BSD 3-Clause License

Copyright (c) 2021, Qing Yu
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:

1. Redistributions of source code must retain the above copyright notice, this
   list of conditions and the following disclaimer.

2. Redistributions in binary form must reproduce the above copyright notice,
   this list of conditions and the following disclaimer in the documentation
   and/or other materials provided with the distribution.

3. Neither the name of the copyright holder nor the names of its
   contributors may be used to endorse or promote products derived from
   this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
'''

import numpy as np
import pandas as pd
from scipy.spatial import cKDTree
import itertools
from operator import itemgetter
import geopandas as gpd
import math
from .coordinates import getdistance


def ckdnearest(dfA_origin,
               dfB_origin,
               Aname=['lon', 'lat'],
               Bname=['lon', 'lat']):
    '''
    Search the nearest points in dfB_origin for dfA_origin,
    and calculate the distance

    Parameters
    -------
    dfA_origin : DataFrame
        DataFrame A
    dfB_origin : DataFrame
        DataFrame B
    Aname : List
        The column of lng and lat in DataFrame A
    Bname : List
        The column of lng and lat in DataFrame A

    Returns
    -------
    gdf : DataFrame
        The output DataFrame
    '''
    if len(dfA_origin) == 0:
        raise Exception('The input DataFrame dfA is empty')
    if len(dfB_origin) == 0:
        raise Exception('The input DataFrame dfB is empty')
    gdA = dfA_origin.copy()
    gdB = dfB_origin.copy()
    from scipy.spatial import cKDTree
    btree = cKDTree(gdB[Bname].values)
    dist, idx = btree.query(gdA[Aname].values, k=1)
    gdA['index'] = idx
    gdB['index'] = range(len(gdB))
    gdf = pd.merge(gdA, gdB, on='index')
    if (Aname[0] == Bname[0]) & (Aname[1] == Bname[1]):
        gdf['dist'] = getdistance(
            gdf[Aname[0]+'_x'],
            gdf[Aname[1]+'_y'],
            gdf[Bname[0]+'_x'],
            gdf[Bname[1]+'_y'])
    else:
        gdf['dist'] = getdistance(
            gdf[Aname[0]],
            gdf[Aname[1]],
            gdf[Bname[0]],
            gdf[Bname[1]])
    return gdf


def ckdnearest_point(gdA, gdB):
    '''
    This method will match the nearest points in gdfB to gdfA,
    and add a new column called dist

    Parameters
    -------
    gdA : GeoDataFrame
        GeoDataFrame A, point geometry

    gdB : GeoDataFrame
        GeoDataFrame B, point geometry

    Returns
    -------
    gdf : DataFrame
        The output DataFrame
    '''
    if len(gdA) == 0:
        raise Exception('The input GeoDataFrame gdfA is empty')
    if len(gdB) == 0:
        raise Exception('The input GeoDataFrame gdfB is empty')
    nA = np.array(list(gdA.geometry.apply(lambda x: (x.x, x.y))))
    nB = np.array(list(gdB.geometry.apply(lambda x: (x.x, x.y))))
    btree = cKDTree(nB)
    dist, idx = btree.query(nA, k=1)
    gdA['dist'] = dist
    gdA['index'] = idx
    gdB['index'] = range(len(gdB))
    gdf = pd.merge(gdA, gdB, on='index')
    return gdf


def ckdnearest_line(gdfA, gdfB):
    '''
    This method will seach from gdfB to find the nearest line to
    the point in gdfA.

    Parameters
    -------
    gdA : GeoDataFrame
        GeoDataFrame A, point geometry

    gdB : GeoDataFrame
        GeoDataFrame B, linestring geometry

    Returns
    -------
    gdf : DataFrame
        Searching the nearset linestring in gdfB for the point in gdfA
    '''
    if len(gdfA) == 0:
        raise Exception('The input GeoDataFrame gdfA is empty')
    if len(gdfB) == 0:
        raise Exception('The input GeoDataFrame gdfB is empty')
    A = np.concatenate(
        [np.array(geom.coords) for geom in gdfA.geometry.to_list()])
    B = [np.array(geom.coords) for geom in gdfB.geometry.to_list()]
    B_ix = tuple(itertools.chain.from_iterable(
        [itertools.repeat(i, x) for i, x in enumerate(list(map(len, B)))]))
    B = np.concatenate(B)
    ckd_tree = cKDTree(B)
    dist, idx = ckd_tree.query(A, k=1)
    idx = itemgetter(*idx)(B_ix)
    gdfA['dist'] = dist
    gdfA['index'] = idx
    gdfB['index'] = range(len(gdfB))
    gdf = pd.merge(gdfA, gdfB, on='index')
    return gdf


def splitline_with_length(Centerline, maxlength=100):
    '''
    The intput is the linestring GeoDataFrame. The splited line’s
    length wull be no longer than maxlength

    Parameters
    -------
    Centerline : GeoDataFrame
        Linestring geometry
    maxlength : number
        The maximum length of the splited line

    Returns
    -------
    splitedline : GeoDataFrame
        Splited line
    '''
    def splitline(route, maxlength):
        routelength = route.length
        from shapely.geometry import LineString
        lss = []
        for k in range(int(routelength/maxlength)+1):
            if k == int(routelength/maxlength):
                lm = routelength
            else:
                lm = (k+1)*maxlength
            a = np.linspace(k*maxlength, lm, 10)
            ls = []
            for line in a:
                ls.append(route.interpolate(line))
            lss.append(LineString(ls))
        lss = gpd.GeoDataFrame(lss, columns=['geometry'])
        return lss
    lsss = []
    for i in range(len(Centerline)):
        route = Centerline['geometry'].iloc[i]
        lss = splitline(route, maxlength)
        lss['id'] = i
        lsss.append(lss)
    lsss = pd.concat(lsss)
    lsss['length'] = lsss.length
    splitedline = lsss
    return splitedline


def merge_polygon(data, col):
    '''
    The input is the GeoDataFrame of polygon geometry, and the col
    name. This function will merge the polygon based on the category
    in the mentioned column

    Parameters
    -------
    data : GeoDataFrame
        The polygon geometry
    col : str
        The column name for indicating category

    Returns
    -------
    data1 : GeoDataFrame
        The merged polygon
    '''
    groupnames = []
    geometries = []
    for i in data[col].drop_duplicates():
        groupnames.append(i)
        geometries.append(data[data[col] == i].unary_union)
    data1 = gpd.GeoDataFrame()
    data1['geometry'] = geometries
    data1[col] = groupnames
    return data1


def polyon_exterior(data, minarea=0):
    '''
    The input is the GeoDataFrame of the polygon geometry. The method
    will construct new polygon by extending the outer boundary of the
    ploygon

    Parameters
    -------
    data : GeoDataFrame
        The polygon geometry
    minarea : number
        The minimum area. Polygon of less area will be removed

    Returns
    -------
    data1 : GeoDataFrame
        The processed polygon
    '''
    data1 = data.copy()

    def polyexterior(p):
        from shapely.geometry import Polygon, MultiPolygon
        if type(p) == MultiPolygon:
            geometries = []
            for i in p:
                poly = Polygon(i.exterior)
                if minarea > 0:
                    if poly.area > minarea:
                        geometries.append(poly)
                else:
                    geometries.append(poly)
            return MultiPolygon(geometries)
        if type(p) == Polygon:
            return Polygon(p.exterior)
    data1['geometry'] = data1['geometry'].apply(polyexterior)
    data1 = data1[-data1['geometry'].is_empty]
    return data1


def ellipse_params(data, col=['lon', 'lat'], confidence=95, epsg=None):
    '''
    confidence ellipse parameter estimation for point data

    Parameters
    -------
    data : DataFrame
        point data
    confidence : number
        confidence level: 99，95 or 90
    epsg : number
        If given, the original coordinates are transformed from WGS84 to
        the given EPSG coordinate system for confidence ellipse parameter
        estimation
    col: List
        Column names, [lon，lat]

    Returns
    -------
    params: List
        Centroid ellipse parameters[pos,width,height,theta,area,oblateness]
        Respectively [Center point coordinates, minor axis, major axis,
        angle, area, oblateness]
    '''
    lon, lat = col
    if confidence == 99:
        nstd = 9.210**0.5
    if confidence == 95:
        nstd = 5.991**0.5
    if confidence == 90:
        nstd = 4.605**0.5
    points = data.copy()
    points = gpd.GeoDataFrame(points)
    points['geometry'] = gpd.points_from_xy(points[lon], points[lat])
    if epsg:
        points.crs = {'init': 'epsg:4326'}
        points = points.to_crs(epsg=epsg)
    point_np = np.array([points.geometry.x, points.geometry.y]).T
    pos = point_np.mean(axis=0)
    cov = np.cov(point_np, rowvar=False)
    vals, vecs = np.linalg.eigh(cov)
    theta = np.degrees(np.arctan2(*vecs[:, 0][::-1]))
    width, height = 2 * nstd * np.sqrt(vals)
    area = width/2*height/2*math.pi
    oblateness = (height-width)/height

    ellip_params = [pos, width, height, theta, area, oblateness]
    return ellip_params


def ellipse_plot(ellip_params, ax, **kwargs):
    '''
    Enter the parameters of the confidence ellipse and plot the confidence
    ellipse

    输入
    -------
    ellip_params : List
        Centroid ellipse parameters[pos,width,height,theta,area,oblateness]
        Respectively[Center point coordinates, minor axis, major axis, angle
        , area, oblateness]

    ax : matplotlib.axes._subplots.AxesSubplot
        Where to plot
    '''
    [pos, width, height, theta, area, alpha] = ellip_params
    from matplotlib.patches import Ellipse
    ellip = Ellipse(xy=pos, width=width, height=height,
                    angle=theta, linestyle='-', **kwargs)
    ax.add_artist(ellip)



'''
BSD 3-Clause License

Copyright (c) 2021, Qing Yu
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:

1. Redistributions of source code must retain the above copyright notice, this
   list of conditions and the following disclaimer.

2. Redistributions in binary form must reproduce the above copyright notice,
   this list of conditions and the following disclaimer in the documentation
   and/or other materials provided with the distribution.

3. Neither the name of the copyright holder nor the names of its
   contributors may be used to endorse or promote products derived from
   this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
'''

import pandas as pd


def bikedata_to_od(data,
                   col=['BIKE_ID', 'DATA_TIME', 'LONGITUDE',
                        'LATITUDE', 'LOCK_STATUS'],
                   startend=None):
    '''
    Input bike-sharing order data (with data only generated
    when the lock is switched on and off), specify the column
    name, and extract the ride and parking information from it

    Parameters
    -------
    data : DataFrame
        Bike-sharing order data
    col : List
        Column names, the order cannot be changed.[‘BIKE_ID’,
        ’DATA_TIME’,’LONGITUDE’,’LATITUDE’,’LOCK_STATUS’]
    startend : List
        The start time and end time of the observation period,
        for example [‘2018-08-27 00:00:00’,’2018-08-28 00:00:00’].
        If it is passed in, the riding and parking situations
        (from the beginning of the observation period to the first
        appearance of the bicycle) and (from the last appearance
        of the bicycle to the end of the observation period) are considered.

    Returns
    -------
    move_data : DataFrame
        Riding data
    stop_data : DataFrame
        Parking data
    '''
    [BIKE_ID, DATA_TIME, LONGITUDE, LATITUDE, LOCK_STATUS] = col
    oddata = data.copy()
    oddata = oddata.sort_values(by=[BIKE_ID, DATA_TIME])
    if startend:
        oddata['tmp_index'] = range(len(oddata))
        # Add records at the beginning of the time period
        data_1 = oddata.copy()
        # After grouping the single vehicle ID, sort it in ascending order
        # according to time to get the rank id
        data_1['rank'] = data_1.groupby(
            BIKE_ID)['tmp_index'].rank(method='first')
        data_1 = data_1[data_1['rank'] == 1]
        # Modify time to the start time of the observation period
        data_1[DATA_TIME] = startend[0]
        data_1[LOCK_STATUS] = 1
        # Add records at the end of the time period
        data_2 = oddata.copy()
        data_2['rank'] = data_2.groupby(BIKE_ID)['tmp_index'].rank(
            ascending=False, method='first')
        data_2 = data_2[data_2['rank'] == 1]
        data_2[DATA_TIME] = startend[1]
        data_2[LOCK_STATUS] = 0
        oddata = pd.concat([oddata, data_1, data_2]).sort_values(
            by=[BIKE_ID, DATA_TIME])
    for i in col:
        oddata[i+'_'] = oddata[i].shift(-1)
    oddata = oddata[oddata[BIKE_ID] == oddata[BIKE_ID+'_']]
    move_data = oddata[(oddata[LOCK_STATUS] == 0) &
                       (oddata[LOCK_STATUS+'_'] == 1)][[
                           BIKE_ID, DATA_TIME, LONGITUDE, LATITUDE,
                           DATA_TIME+'_', LONGITUDE+'_', LATITUDE+'_']]
    move_data.columns = [BIKE_ID, 'stime',
                         'slon', 'slat', 'etime', 'elon', 'elat']
    stop_data = oddata[(oddata[LOCK_STATUS] == 1) &
                       (oddata[LOCK_STATUS+'_'] == 0)][[
                           BIKE_ID, DATA_TIME, LONGITUDE, LATITUDE,
                           DATA_TIME+'_', LONGITUDE+'_', LATITUDE+'_']]
    stop_data.columns = [BIKE_ID, 'stime',
                         'slon', 'slat', 'etime', 'elon', 'elat']
    return move_data, stop_data



import setuptools

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

setuptools.setup(
    name="transbigdata",
    version="0.4.10",
    author="Qing Yu",
    author_email="qingyu0815@foxmail.com",
    description="A Python package developed for transportation spatio-temporal big data processing and analysis.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    license="BSD",
    url="https://github.com/ni1o1/transbigdata",
    project_urls={
        "Bug Tracker": "https://github.com/ni1o1/transbigdata/issues",
    },
    install_requires=[
        "numpy", "pandas", "shapely", "geopandas", "scipy", "matplotlib"
    ],
    classifiers=[
        "Operating System :: OS Independent",
        "Topic :: Text Processing :: Indexing",
        "Topic :: Utilities",
        "Topic :: Software Development :: Libraries :: Python Modules",
        "License :: OSI Approved :: BSD License",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
    ],
    package_dir={'transbigdata': 'src/transbigdata'},
    packages=['transbigdata'],
    python_requires=">=3.6",
)



'''
BSD 3-Clause License

Copyright (c) 2021, Qing Yu
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:

1. Redistributions of source code must retain the above copyright notice, this
   list of conditions and the following disclaimer.

2. Redistributions in binary form must reproduce the above copyright notice,
   this list of conditions and the following disclaimer in the documentation
   and/or other materials provided with the distribution.

3. Neither the name of the copyright holder nor the names of its
   contributors may be used to endorse or promote products derived from
   this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
'''

import pandas as pd


def clean_taxi_status(data, col=['VehicleNum', 'Time', 'OpenStatus'],
                      timelimit=None):
    '''
    Deletes records of instantaneous changes in passenger carrying status from
    taxi data. These abnormal records can affect travel order judgments.
    Judgement method: If the passenger status of the previous record and the
    next record are different from this record for the same vehicle, then this
    record should be deleted.

    Parameters
    -------
    data : DataFrame
        Data
    col : List
        Column names, in the order of [‘VehicleNum’, ‘Time’, ‘OpenStatus’]
    timelimit : number
        Optional, in seconds. If the time between the previous record and the
        next record is less than the time threshold, then it will be deleted

    Returns
    -------
    data1 : DataFrame
        Cleaned data
    '''
    data1 = data.copy()
    [VehicleNum, Time, OpenStatus] = col
    if timelimit:
        data1[Time] = pd.to_datetime(data1[Time])
        data1 = data1[
            -((data1[OpenStatus].shift(-1) == data1[OpenStatus].shift()) &
              (data1[OpenStatus] != data1[OpenStatus].shift()) &
              (data1[VehicleNum].shift(-1) == data1[VehicleNum].shift()) &
              (data1[VehicleNum] == data1[VehicleNum].shift()) &
              ((data1[Time].shift(-1) - data1[Time].shift()
                ).dt.total_seconds() <= timelimit)
              )]
    else:
        data1 = data1[
            -((data1[OpenStatus].shift(-1) == data1[OpenStatus].shift()) &
              (data1[OpenStatus] != data1[OpenStatus].shift()) &
              (data1[VehicleNum].shift(-1) == data1[VehicleNum].shift()) &
              (data1[VehicleNum] == data1[VehicleNum].shift()))]
    return data1


def taxigps_to_od(data,
                  col=['VehicleNum', 'Stime', 'Lng', 'Lat', 'OpenStatus']):
    '''
    Input Taxi GPS data, extract OD information

    Parameters
    -------
    data : DataFrame
        Taxi GPS data
    col : List
        Column names in the data, need to be in order [vehicle id, time,
        longitude, latitude, passenger status]

    Returns
    -------
    oddata : DataFrame
        OD information
    '''
    [VehicleNum, Stime, Lng, Lat, OpenStatus] = col
    data1 = data[col]
    data1 = data1.sort_values(by=[VehicleNum, Stime])
    data1['StatusChange'] = data1[OpenStatus] - data1[OpenStatus].shift()
    oddata = data1[((data1['StatusChange'] == -1) |
                   (data1['StatusChange'] == 1)) &
                   (data1[VehicleNum].shift() == data1[VehicleNum])]
    oddata = oddata.drop([OpenStatus], axis=1)
    oddata.columns = [VehicleNum, 'stime', 'slon', 'slat', 'StatusChange']
    oddata['etime'] = oddata['stime'].shift(-1)
    oddata['elon'] = oddata['slon'].shift(-1)
    oddata['elat'] = oddata['slat'].shift(-1)
    oddata = oddata[(oddata['StatusChange'] == 1) &
                    (oddata[VehicleNum] == oddata[VehicleNum].shift(-1))]
    oddata = oddata.drop('StatusChange', axis=1)
    oddata['ID'] = range(len(oddata))
    return oddata


def taxigps_traj_point(data, oddata,
                       col=['Vehicleid', 'Time', 'Lng', 'Lat', 'OpenStatus']):
    '''
    Input Taxi data and OD data to extract the trajectory points for delivery
    and idle trips

    Parameters
    -------
    data : DataFrame
        Taxi GPS data, field name specified by col variable
    oddata : DataFrame
        Taxi OD data
    col : List
        Column names, need to be in order [vehicle id, time, longitude,
        latitude, passenger status]

    Returns
    -------
    data_deliver : DataFrame
        Trajectory points for delivery trips
    data_idle : DataFrame
        Trajectory points for idle trips
    '''
    VehicleNum, Time, Lng, Lat, OpenStatus = col
    oddata1 = oddata.copy()
    odata = oddata1[[VehicleNum, 'stime', 'slon', 'slat', 'ID']].copy()
    odata.columns = [VehicleNum, Time, Lng, Lat, 'ID']
    odata.loc[:, 'flag'] = 1
    odata.loc[:, OpenStatus] = -1
    ddata = oddata1[[VehicleNum, 'etime', 'elon', 'elat', 'ID']].copy()
    ddata.columns = [VehicleNum, Time, Lng, Lat, 'ID']
    ddata.loc[:, 'flag'] = -1
    ddata.loc[:, OpenStatus] = -1
    data1 = pd.concat([data, odata, ddata])
    data1 = data1.sort_values(by=[VehicleNum, Time, OpenStatus])
    data1['flag'] = data1['flag'].fillna(0)
    data1['flag'] = data1.groupby(VehicleNum)['flag'].cumsum()
    data1['ID'] = data1['ID'].ffill()
    data_deliver = data1[(data1['flag'] == 1) &
                         (-data1['ID'].isnull()) & (data1[OpenStatus] != -1)]
    data_idle = data1[(data1['flag'] == 0) &
                      (-data1['ID'].isnull()) & (data1[OpenStatus] != -1)]
    return data_deliver, data_idle



'''
BSD 3-Clause License

Copyright (c) 2021, Qing Yu
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:

1. Redistributions of source code must retain the above copyright notice, this
   list of conditions and the following disclaimer.

2. Redistributions in binary form must reproduce the above copyright notice,
   this list of conditions and the following disclaimer in the documentation
   and/or other materials provided with the distribution.

3. Neither the name of the copyright holder nor the names of its
   contributors may be used to endorse or promote products derived from
   this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
'''

import geopandas as gpd
import pandas as pd
from .grids import (
    GPS_to_grid,
    area_to_params,
    grid_to_centre,
    grid_to_polygon
)
from .coordinates import getdistance


def clean_same(data, col=['VehicleNum', 'Time', 'Lng', 'Lat']):
    '''
    Delete the data with the same information as the data before and
    after to reduce the amount of data. For example, if several consecutive
    data of an individual have the same information except for the time,
    only the first and last two data can be kept

    Parameters
    -------
    data : DataFrame
        Data
    col : List
        The column name, in the order of [‘Vehicleid, Time’]. It will sort
        by time, and then determine the information of other columns besides
        the time

    Returns
    -------
    data1 : DataFrame
        Cleaned data
    '''
    [VehicleNum, Time, Lng, Lat] = col[:4]
    extra = col[4:]
    data1 = data.copy()
    data1 = data1.drop_duplicates(subset=[VehicleNum, Time])
    data1 = data1.sort_values(by=[VehicleNum, Time])
    data1['issame'] = 0
    for i in [VehicleNum, Lng, Lat]+extra:
        data1['issame'] += (data1[i].shift() == data1[i]
                            ) & (data1[i].shift(-1) == data1[i])
    data1 = data1[-(data1['issame'] == len([VehicleNum, Lng, Lat]+extra))]
    data1 = data1.drop('issame', axis=1)
    return data1


def clean_drift(data, col=['VehicleNum', 'Time', 'Lng', 'Lat'],
                speedlimit=80, dislimit=1000):
    '''
    Delete the drift data. The select principle is that: if the speed of a
    trajectory point is larger than the speed limit with before and after
    points, while the speed between the before and after data is less than
    the speedlimit. The time column in the input data is calculated more
    efficiently if it is in datetime format.

    Parameters
    -------
    data : DataFrame
        Data
    col : List
        Column names, in the order of [‘VehicleNum’, ‘Time’, ‘Lng’, ‘Lat’]
    speedlimit : number
        Speed limitation

    Returns
    -------
    data1 : DataFrame
        Cleaned data
    '''
    [VehicleNum, Time, Lng, Lat] = col
    data1 = data.copy()
    data1 = data1.drop_duplicates(subset=[VehicleNum, Time])
    data1[Time+'_dt'] = pd.to_datetime(data1[Time])
    data1 = data1.sort_values(by=[VehicleNum, Time])
    for i in [VehicleNum, Lng, Lat, Time+'_dt']:
        data1[i+'_pre'] = data1[i].shift()
        data1[i+'_next'] = data1[i].shift(-1)
    data1['dis_pre'] = getdistance(
        data1[Lng],
        data1[Lat],
        data1[Lng+'_pre'],
        data1[Lat+'_pre'])
    data1['dis_next'] = getdistance(
        data1[Lng],
        data1[Lat],
        data1[Lng+'_next'],
        data1[Lat+'_next'])
    data1['dis_prenext'] = getdistance(
        data1[Lng+'_pre'],
        data1[Lat+'_pre'],
        data1[Lng+'_next'],
        data1[Lat+'_next'])
    data1['timegap_pre'] = data1[Time+'_dt'] - data1[Time+'_dt_pre']
    data1['timegap_next'] = data1[Time+'_dt_next'] - data1[Time+'_dt']
    data1['timegap_prenext'] = data1[Time+'_dt_next'] - data1[Time+'_dt_pre']
    data1['speed_pre'] = data1['dis_pre'] / \
        data1['timegap_pre'].dt.total_seconds()*3.6
    data1['speed_next'] = data1['dis_next'] / \
        data1['timegap_next'].dt.total_seconds()*3.6
    data1['speed_prenext'] = data1['dis_prenext'] / \
        data1['timegap_prenext'].dt.total_seconds()*3.6
    if speedlimit:
        data1 = data1[
            -((data1[VehicleNum+'_pre'] == data1[VehicleNum]) &
              (data1[VehicleNum+'_next'] == data1[VehicleNum]) &
                (data1['speed_pre'] > speedlimit) &
              (data1['speed_next'] > speedlimit) &
              (data1['speed_prenext'] < speedlimit))]
    if dislimit:
        data1 = data1[
            -((data1[VehicleNum+'_pre'] == data1[VehicleNum]) &
              (data1[VehicleNum+'_next'] == data1[VehicleNum]) &
              (data1['dis_pre'] > dislimit) &
              (data1['dis_next'] > dislimit) &
              (data1['dis_prenext'] < dislimit))]
    data1 = data1[data.columns]
    return data1


def clean_outofbounds(data, bounds, col=['Lng', 'Lat']):
    '''
    The input is the latitude and longitude coordinates of the lower
    left and upper right of the study area and exclude data that are
    outside the study area

    Parameters
    -------
    data : DataFrame
        Data
    bounds : List
        Latitude and longitude of the lower left and upper right of
        the study area, in the order of [lon1, lat1, lon2, lat2]
    col : List
        Column name of longitude and latitude

    Returns
    -------
    data1 : DataFrame
        Data within the scope of the study
    '''
    lon1, lat1, lon2, lat2 = bounds
    if (lon1 > lon2) | (lat1 > lat2) | (abs(lat1) > 90) | (
            abs(lon1) > 180) | (abs(lat2) > 90) | (abs(lon2) > 180):
        raise Exception(
            'Bounds error. The input bounds should be in the order \
of [lon1,lat1,lon2,lat2]. (lon1,lat1) is the lower left corner and \
(lon2,lat2) is the upper right corner.')
    Lng, Lat = col
    data1 = data.copy()
    data1 = data1[(data1[Lng] > bounds[0]) & (data1[Lng] < bounds[2]) & (
        data1[Lat] > bounds[1]) & (data1[Lat] < bounds[3])]
    return data1


def clean_outofshape(data, shape, col=['Lng', 'Lat'], accuracy=500):
    '''
    Input the GeoDataFrame of the study area and exclude the data beyond
    the study area

    Parameters
    -------
    data : DataFrame
        Data
    shape : GeoDataFrame
        The GeoDataFrame of the study area
    col : List
        Column name of longitude and latitude
    accuracy : number
        The size of grid. The principle is to do the data gridding first
        and then do the data cleaning. The smaller the size is, the higher
        accuracy it has

    Returns
    -------
    data1 : DataFrame
        Data within the scope of the study
    '''
    Lng, Lat = col
    shape_unary = shape.unary_union
    bounds = shape_unary.bounds
    params = area_to_params(bounds, accuracy)
    data1 = data.copy()
    data1['LONCOL'], data1['LATCOL'] = GPS_to_grid(
        data1[Lng], data1[Lat], params)
    data1_gdf = data1[['LONCOL', 'LATCOL']].drop_duplicates()
    data1_gdf['geometry'] = grid_to_polygon(
        [data1_gdf['LONCOL'], data1_gdf['LATCOL']], params)
    data1_gdf = gpd.GeoDataFrame(data1_gdf)
    data1_gdf = data1_gdf[data1_gdf.intersects(shape_unary)]
    data1 = pd.merge(data1, data1_gdf[['LONCOL', 'LATCOL']]).drop(
        ['LONCOL', 'LATCOL'], axis=1)
    return data1


def clean_traj(data, col=['uid', 'str_time', 'lon', 'lat'], tripgap=1800,
               disgap=50000, speedlimit=80):
    '''
    A combo for trajectory data cleaning, including defining the the time
    length threshold considered as a new trip, and the distance threshold
    considered as a new trip

    Parameters
    -------
    data : DataFrame
        Trajectory data
    col : List
        Column names, in the order of [‘VehicleNum’, ‘Time’, ‘Lng’, ‘Lat’]
    tripgap : number
        The time length threshold considered as a new trip
    disgap : number
        The distance threshold considered as a new trip
    speedlimit : number
        Speed limit

    Returns
    -------
    data1 : DataFrame
        Cleaned data
    '''
    uid, timecol, lon, lat = col
    data[timecol] = pd.to_datetime(data[timecol])
    data = data.sort_values(by=[uid, timecol])
    cols = []
    for i in data.columns:
        if i not in [uid, timecol, lon, lat]:
            cols.append(i)
    data = clean_same(data, col=[uid, timecol, lon, lat]+cols)
    data = clean_drift(data, col=[uid, timecol, lon, lat],
                       speedlimit=speedlimit)
    data = id_reindex(data, uid, timecol=timecol, timegap=tripgap)
    data = data.rename(columns={uid+'_new': 'tripid'})
    data = id_reindex_disgap(
        data, col=['tripid', lon, lat], disgap=disgap, suffix='')
    data1 = data.copy()
    data1['lon1'] = data1[lon].shift(-1)
    data1['lat1'] = data1[lat].shift(-1)
    data1['tripid1'] = data1['tripid'].shift(-1)
    data1 = data1[data1['tripid'] == data1['tripid1']]
    data1['dis'] = getdistance(
        data1[lon], data1[lat], data1['lon1'], data1['lat1'])
    a = data1.groupby(['tripid'])['dis'].sum()
    a = a[-(a < 50)].reset_index()['tripid']
    data = pd.merge(data, a)
    data = data.drop('tripid', axis=1)
    data = id_reindex(data, uid, timecol=timecol, timegap=tripgap)
    data = data.rename(columns={uid+'_new': 'tripid'})
    data = id_reindex_disgap(
        data, col=['tripid', lon, lat], disgap=disgap, suffix='')
    return data


def dataagg(data, shape, col=['Lng', 'Lat', 'count'], accuracy=500):
    '''
    Aggregate data to traffic zone

    Parameters
    -------
    data : DataFrame
        The origin DataFrame
    shape : GeoDataFrame
        The shape of the traffic zone
    col : List
        You can either choose to input two columns, i.e., [‘Lng’,’Lat’], or
        to input three columns, i.e., [‘Lng’,’Lat’,’count’]”, where count
        means the points count
    accuracy : number
        The idea is to first implement data gridding and then the aggregation.
        Here, the grid size will be determined. The less the size is, the
        higher the accuracy will have.

    Returns
    -------
    aggresult : GeoDataFrame
        Traffic zone. The count column is the output result
    data1 : DataFrame
        The zone-matched data
    '''
    if len(col) == 2:
        Lng, Lat = col
        aggcol = None
    else:
        Lng, Lat, aggcol = col
    shape['index'] = range(len(shape))
    shape_unary = shape.unary_union
    bounds = shape_unary.bounds
    params = area_to_params(bounds, accuracy)
    data1 = data.copy()
    data1['LONCOL'], data1['LATCOL'] = GPS_to_grid(
        data1[Lng], data1[Lat], params)
    data1_gdf = data1[['LONCOL', 'LATCOL']].drop_duplicates()
    data1_gdf['geometry'] = gpd.points_from_xy(
        *grid_to_centre([data1_gdf['LONCOL'], data1_gdf['LATCOL']], params))
    data1_gdf = gpd.GeoDataFrame(data1_gdf)
    data1_gdf = gpd.sjoin(data1_gdf, shape, how='left')
    data1 = pd.merge(data1, data1_gdf).drop(['LONCOL', 'LATCOL'], axis=1)
    if aggcol:
        aggresult = pd.merge(shape, data1.groupby('index')[
                             aggcol].sum().reset_index()).drop('index', axis=1)
    else:
        data1['_'] = 1
        aggresult = pd.merge(shape, data1.groupby('index')['_'].sum().rename(
            'count').reset_index()).drop('index', axis=1)
        data1 = data1.drop('_', axis=1)
    data1 = data1.drop('index', axis=1)
    return aggresult, data1


def id_reindex_disgap(data, col=['uid', 'lon', 'lat'], disgap=1000,
                      suffix='_new'):
    '''
    Renumber the ID columns of the data，If two adjacent records exceed the
    distance, the number is the new ID

    Parameters
    -------
    data : DataFrame
        Data
    col : str
        Name of the ID column to be re-indexed
    disgap : number
        If two adjacent records exceed this distance, the number is the
        new ID
    suffix : str
        The suffix of the new column. When set to False, the former column
        will be replaced

    Returns
    -------
    data1 : DataFrame
        Renumbered data
    '''
    uid, lon, lat = col
    data1 = data.copy()
    data1[uid+suffix] = (
        (data1[uid].shift() != data1[uid]) |
        (getdistance(data1[lon],
                     data1[lat],
                     data1[lon].shift(),
                     data1[lat].shift()) > disgap)).astype(int).cumsum()-1
    a = data1.groupby([uid+suffix])[lon].count()
    data1 = pd.merge(data1, a[a > 1].reset_index()[[uid+suffix]])
    return data1


def id_reindex(data, col, new=False, timegap=None, timecol=None,
               suffix='_new', sample=None):
    '''
    Renumber the ID columns of the data

    Parameters
    -------
    data : DataFrame
        Data
    col : str
        Name of the ID column to be re-indexed
    new : bool
        False: the new number of the same ID will be the same index;
        True: according to the order of the table, the origin ID appears
        again with different index
    timegap : number
        If an individual does not appear for a period of time (timegap is
        the time threshold), it is numbered as a new individual. This parameter
        should be set with timecol to take effect.
    timecol : str
        The column name of time, it should be set with timegap to take effect
    suffix : str
        The suffix of the new column. When set to False, the former column will
        be replaced
    sample : int (optional)
        To desampling the data

    Returns
    -------
    data1 : DataFrame
        Renumbered data
    '''
    if not suffix:
        suffix = ''
    data1 = data.copy()
    if new:
        data1[col+suffix] = data1[col] != data1[col].shift()
        data1[col+suffix] = data1[col+suffix].cumsum()-1
    else:
        tmp = data1[[col]].drop_duplicates()
        tmp[col+'_'] = range(len(tmp))
        data1 = pd.merge(data1, tmp, on=col)
        data1[col+suffix] = data1[col+'_']
        if suffix != '_':
            data1 = data1.drop(col+'_', axis=1)
    if (timegap is not None) & (timecol is not None):
        data1[timecol] = pd.to_datetime(data1[timecol])
        data1 = data1.sort_values(by=[col+suffix, timecol])
        data1[col+suffix] = (
            (data1[col+suffix].shift() != data1[col+suffix]) |
            ((data1[timecol]-data1[timecol].shift()).dt.total_seconds()
             > timegap)).astype(int).cumsum()-1

    if sample:
        tmp = data1[col+suffix].drop_duplicates().sample(sample)
        data1 = pd.merge(data1, tmp)
    return data1



'''
BSD 3-Clause License

Copyright (c) 2021, Qing Yu
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:

1. Redistributions of source code must retain the above copyright notice, this
   list of conditions and the following disclaimer.

2. Redistributions in binary form must reproduce the above copyright notice,
   this list of conditions and the following disclaimer in the documentation
   and/or other materials provided with the distribution.

3. Neither the name of the copyright holder nor the names of its
   contributors may be used to endorse or promote products derived from
   this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
'''

import pandas as pd
import numpy as np
import geopandas as gpd
from shapely.geometry import Polygon, LineString
import urllib.request
from urllib import parse
import urllib
import json
import re
from .plotmap import read_mapboxtoken
from .coordinates import (
    gcj02towgs84,
    bd09towgs84,
    bd09mctobd09,
    wgs84togcj02
)


def getadmin(keyword, ak, jscode='', subdistricts=False):
    '''
    Input the keyword and the Amap ak. The output is the GIS file of
    the administrative boundary (Only in China)

    Parameters
    -------
    keywords : str
        The keyword. It might be the city name such as Shengzheng, or
        the administrative code such as 440500
    ak : str
        Amap accesstoken
    jscode : jscode
        Amap safty code
    subdistricts : bool
        Whether to output the information of the administrative district
        boundary

    Returns
    -------
    admin : GeoDataFrame
        Administrative district(WGS84)
    districts : DataFrame
        The information of subdistricts. This can be used to further get
        the boundary of lower level districts
    '''

    # API url
    url = 'https://restapi.amap.com/v3/config/district?'
    # Condition
    dict1 = {
        'subdistrict': '3',
        'showbiz': 'false',
        'extensions': 'all',
        'key': ak,
        'jscode': jscode,
        's': 'rsv3',
        'output': 'json',
        'level': 'district',
        'keywords': keyword,
        'platform': 'JS',
        'logversion': '2.0',
        'sdkversion': '1.4.10'
    }
    url_data = parse.urlencode(dict1)
    url = url+url_data
    request = urllib.request.Request(url)
    response = urllib.request.urlopen(request)
    webpage = response.read()
    result = json.loads(webpage.decode('utf8', 'ignore'))
    # Organize Data
    datas = []
    k = 0
    polyline = result['districts'][k]['polyline']
    polyline1 = polyline.split('|')
    res = []
    for polyline2 in polyline1:
        polyline2 = polyline2.split(';')
        p = []
        for i in polyline2:
            a, b = i.split(',')
            p.append([float(a), float(b)])
        x = pd.DataFrame(p)
        x[0], x[1] = gcj02towgs84(x[0], x[1])
        p = x.values
        res.append(Polygon(p))
    data = pd.DataFrame()
    data1 = pd.DataFrame()
    data1['geometry'] = res
    data1 = gpd.GeoDataFrame(data1)
    poly = data1.unary_union
    data['geometry'] = [poly]
    try:
        data['citycode'] = result['districts'][k]['citycode']
    except Exception:
        pass
    try:
        data['adcode'] = result['districts'][k]['adcode']
    except Exception:
        pass
    try:
        data['name'] = result['districts'][k]['name']
    except Exception:
        pass
    try:
        data['level'] = result['districts'][k]['level']
    except Exception:
        pass
    try:
        data['center'] = result['districts'][k]['center']
    except Exception:
        pass
    datas.append(data)
    datas = pd.concat(datas)
    admin = gpd.GeoDataFrame(datas)
    if subdistricts:
        districts = result['districts'][k]['districts']
        districts = pd.DataFrame(districts)
        return admin, districts
    else:
        return admin


def getbusdata(city, keywords, accurate=True):
    '''
    Obtain the geographic information of the bus station and bus line from
    the map service (Only in China)

    Parameters
    -------
    city : str
        city name
    keywords : list
        Keyword, the line name
    accurate : bool
        Accurate matching

    Returns
    -------
    data : GeoDataFrame
        The generated bus line(WGS84)
    stop : GeoDataFrame
        The generated bus station(WGS84)
    '''
    def getlineuid(keyword, c, acc=True):
        url = 'http://map.baidu.com/?qt=s&wd=' + \
            urllib.parse.quote(keyword)+'&c='+c+'&from=webmap'
        response1 = urllib.request.urlopen(url)
        searchinfo = json.loads(response1.read().decode('utf8'))
        try:
            res = pd.DataFrame(searchinfo['content'])
            if acc:
                res = list(res[(res['geo_type'] == 1) &
                           (res['acc_flag'] == 1)]['uid'])
                return res
            else:
                res = list(res[res['geo_type'] == 1]['uid'])
                return res
        except Exception:
            return []

    def getcitycode(c):
        url = 'http://map.baidu.com/?qt=s&wd='+urllib.parse.quote(c)
        response1 = urllib.request.urlopen(url, timeout=60)
        searchinfo = json.loads(response1.read().decode('utf8'))
        return str(searchinfo['content']['code'])

    def getlinegeo(uid, c):
        url = 'http://map.baidu.com/?qt=bsl&uid='+uid+'&c='+c+"&auth=1"
        response = urllib.request.urlopen(url, timeout=60)
        searchinfo = json.loads(response.read().decode('utf8'))
        linename = searchinfo['content'][0]['name']
        stations = searchinfo['content'][0]['stations']
        geo = searchinfo['content'][0]['geo'].split('|')[2][:-1].split(',')
        stationgeo = []
        stationnames = []
        for station in stations:
            stationname = station['name']
            coo = station['geo'].split(';')[1].split('|')[0]
            stationnames.append(stationname)
            stationgeo.append(coo)
        coo = []
        t = 0
        cood = ''
        for each in geo:
            t += 1
            cood += each + ','
            if t == 2:
                t = 0
                coo.append(cood[:-1])
                cood = ''

        def coodconvert(coo):
            coo = pd.DataFrame(
                list(pd.DataFrame(coo)[0].str.split(','))).astype(float)
            coo[0], coo[1] = bd09mctobd09(coo[0], coo[1])
            return list(coo[0].astype(str)+','+coo[1].astype(str))
        return (
            linename,
            coodconvert(coo),
            stationnames,
            coodconvert(stationgeo)
        )
    print('Obtaining city id:', city, end='')
    linenames = []
    lines = []
    c = getcitycode(city)
    print('success')
    stop = []
    uids = []
    if type(keywords) != list:
        keywords = [str(keywords)]
    for keyword in keywords:
        print(keyword)
        for uid in getlineuid(keyword, c, accurate):
            if uid not in uids:
                try:
                    linename, coo, stationnames, stationgeo = getlinegeo(
                        uid, c)
                    coo = pd.DataFrame(
                        list(pd.DataFrame(coo)[0].str.split(',')))
                    coo[0], coo[1] = bd09towgs84(coo[0], coo[1])
                    line = LineString(coo.values)
                    linenames.append(linename)
                    lines.append(line)
                    stops = pd.DataFrame({'stationnames': stationnames})
                    stops['linename'] = linename
                    stops['geo'] = stationgeo
                    stops['lon'] = stops['geo'].apply(
                        lambda row: row.split(',')[0])
                    stops['lat'] = stops['geo'].apply(
                        lambda row: row.split(',')[1])
                    stop.append(stops)
                    print(linename+' success')
                    uids.append(uid)
                except Exception:
                    pass
    if len(stop) == 0:
        print('No such busline')
        return gpd.GeoDataFrame(), gpd.GeoDataFrame()
    data = gpd.GeoDataFrame()
    data['linename'] = linenames
    data['geometry'] = lines
    data['city'] = city
    stop = pd.concat(stop)
    stop['lon'], stop['lat'] = bd09towgs84(stop['lon'], stop['lat'])
    stop['geometry'] = gpd.points_from_xy(stop['lon'], stop['lat'])
    stop = stop.drop('geo', axis=1)
    stop = gpd.GeoDataFrame(stop)
    data['line'] = data['linename'].str.split('(').apply(lambda r: r[0])
    stop['line'] = stop['linename'].str.split('(').apply(lambda r: r[0])
    stop['id'] = range(len(stop))
    stop['id'] = stop.groupby('linename')['id'].rank()
    data = data.drop_duplicates(subset=['linename'])
    stop = stop.drop_duplicates(subset=['linename', 'stationnames'])
    return data, stop


def get_isochrone_amap(lon, lat, reachtime, ak, jscode='', mode=2):
    '''
    Obtain the isochrone from Amap reachcricle

    Parameters
    -------
    lon : float
        Longitude of the start point(WGS84)
    lat : float
        Latitude of the start point(WGS84)
    reachtime : number
        Reachtime of the isochrone
    ak : str
        Amap access token
    jscode : jscode
        Amap safty code
    mode : int or str
        Travel mode, should be 0(bus), 1(subway), 2(bus+subway)

    Returns
    -------
    isochrone : GeoDataFrame
        The isochrone GeoDataFrame(WGS84)
    '''
    strategy = str(mode)
    if strategy not in ['0', '1', '2']:
        raise ValueError(
            'Travel mode, should be 0(bus), 1(subway), 2(bus+subway)')
    lon, lat = wgs84togcj02(lon, lat)
    url = 'http://restapi.amap.com/v3/direction/reachcircle?'
    dict1 = {
        'key': ak,
        'jscode': jscode,
        'location': str(round(float(lon), 6))+','+str(round(float(lat), 6)),
        'time': reachtime,
        'output': 'json',
        's': 'rsv3',
        'extensions': 'all',
        'strategy': str(strategy)
    }
    url_data = parse.urlencode(dict1)
    url = url+url_data
    request = urllib.request.Request(url)
    response = urllib.request.urlopen(request, timeout=10)
    webpage = response.read()
    result = json.loads(webpage.decode('utf8', 'ignore'))
    P_all = []
    for each in result['polylines']:
        data = each['outer']
        ll = re.split('[,;]', data)
        ll = np.reshape(ll, (int(len(ll)/2), 2))
        l2 = []
        for i in ll:
            l2.append((gcj02towgs84(float(i[0]), float(i[1]))))
        P_all.append(Polygon(l2))

    a = gpd.GeoSeries(P_all)
    g = gpd.GeoSeries(a[a.is_valid].unary_union)
    g = gpd.GeoDataFrame(g, columns=['geometry'])
    g['lon'] = lon
    g['lat'] = lat
    g['reachtime'] = reachtime
    return g


def get_isochrone_mapbox(lon, lat, reachtime, access_token='auto',
                         mode='driving'):
    '''
    Obtain the isochrone from mapbox isochrone

    Parameters
    -------
    lon : float
        Longitude of the start point(WGS84)
    lat : float
        Latitude of the start point(WGS84)
    reachtime : number
        Reachtime of the isochrone
    access_token : str
        Mapbox access token, if `auto` it will use the preset access token
    mode : bool
        Travel mode, should be `driving`, `walking` or `cycling`

    Returns
    -------
    isochrone : GeoDataFrame
        The isochrone GeoDataFrame(WGS84)
    '''
    if access_token == 'auto':
        access_token = read_mapboxtoken()
    if mode not in ['driving', 'walking', 'cycling']:
        raise ValueError(
            'Travel mode should be `driving`, `walking` or `cycling`')
    url = 'https://api.mapbox.com/isochrone/v1/mapbox/'+mode+'/' +\
        str(lon)+','+str(lat)+'?contours_minutes='+str(reachtime) +\
        '&polygons=true&access_token='+access_token
    request = urllib.request.Request(url)
    response = urllib.request.urlopen(request, timeout=1)
    webpage = response.read()
    result = webpage.decode('utf8', 'ignore')
    isochrone = gpd.GeoDataFrame.from_features(json.loads(result))
    isochrone['lon'] = lon
    isochrone['lat'] = lat
    isochrone['reachtime'] = reachtime
    return isochrone[['lon', 'lat', 'reachtime', 'geometry']]



'''
BSD 3-Clause License

Copyright (c) 2021, Qing Yu
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:

1. Redistributions of source code must retain the above copyright notice, this
   list of conditions and the following disclaimer.

2. Redistributions in binary form must reproduce the above copyright notice,
   this list of conditions and the following disclaimer in the documentation
   and/or other materials provided with the distribution.

3. Neither the name of the copyright holder nor the names of its
   contributors may be used to endorse or promote products derived from
   this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
'''

import pandas as pd


def sample_duration(data, col=['Vehicleid', 'Time']):
    '''
    Calculate the data sampling interval.

    Parameters
    -------
    data : DataFrame
        Data
    col : List
        The column name, in the order of [‘Vehicleid’, ‘Time’]

    Returns
    -------
    sample_duration : DataFrame
        A Series with the column name duration, the content is the sampling
        interval of the data, in seconds
    '''
    [Vehicleid, Time] = col
    data1 = data.copy()
    data1[Time] = pd.to_datetime(data1[Time])
    data1 = data1.sort_values(by=[Vehicleid, Time])
    data1[Vehicleid+'1'] = data1[Vehicleid].shift(-1)
    data1[Time+'1'] = data1[Time].shift(-1)
    data1['duration'] = (data1[Time+'1']-data1[Time]).dt.total_seconds()
    data1 = data1[data1[Vehicleid+'1'] == data1[Vehicleid]]
    sample_duration = data1[['duration']]
    return sample_duration


def data_summary(data, col=['Vehicleid', 'Time'], show_sample_duration=False,
                 roundnum=4):
    '''
    Output the general information of the dataset.

    Parameters
    -------
    data : DataFrame
        The trajectory points data
    col : List
        The column name, in the order of [‘Vehicleid’, ‘Time’]
    show_sample_duration : bool
        Whether to output individual sampling interval
    roundnum : number
        Number of decimal places
    '''
    [Vehicleid, Time] = col
    print('Amount of data')
    print('-----------------')
    print('Total number of data items: ', len(data))
    Vehicleid_count = data[Vehicleid].value_counts()
    print('Total number of individuals: ', len(Vehicleid_count))
    print('Data volume of individuals(Mean): ',
          round(Vehicleid_count.mean(), roundnum))
    print('Data volume of individuals(Upper quartile): ',
          round(Vehicleid_count.quantile(0.75), roundnum))
    print('Data volume of individuals(Median): ', round(
        Vehicleid_count.quantile(0.5), roundnum))
    print('Data volume of individuals(Lower quartile): ',
          round(Vehicleid_count.quantile(0.25), roundnum))
    print('')
    print('Data time period')
    print('-----------------')
    print('Start time: ', data[Time].min())
    print('End time: ', data[Time].max())
    print('')
    if show_sample_duration:
        sd = sample_duration(data, col=[Vehicleid, Time])
        print('Sampling interval')
        print('-----------------')
        print('Mean: ', round(sd['duration'].mean(), roundnum), 's')
        print('Upper quartile: ', round(
            sd['duration'].quantile(0.75), roundnum), 's')
        print('Median: ', round(sd['duration'].quantile(0.5), roundnum), 's')
        print('Lower quartile: ', round(
            sd['duration'].quantile(0.25), roundnum), 's')
